import tempfile
from queue import Queue

import mock
import pytest

from tron.bin import recover_batch
from tron.bin.action_runner import StatusFile


@pytest.fixture
def mock_file():
    f = tempfile.NamedTemporaryFile()
    yield f.name
    f.close()


@mock.patch.object(recover_batch, 'reactor')
@mock.patch('tron.bin.recover_batch.get_exit_code', autospec=True)
@pytest.mark.parametrize('exit_code,error_msg,should_stop', [
    (1, 'failed', True),
    (None, None, False),
])
def test_notify(mock_get_exit_code, mock_reactor, exit_code, error_msg, should_stop):
    mock_get_exit_code.return_value = exit_code, error_msg
    queue = Queue()
    path = mock.Mock()
    recover_batch.notify(queue, 'some_ignored', path, 'mask')
    if should_stop:
        assert mock_reactor.stop.call_count == 1
        assert queue.get_nowait() == (exit_code, error_msg)
    else:
        assert mock_reactor.stop.call_count == 0
        assert queue.empty()


@mock.patch('tron.bin.recover_batch.psutil.pid_exists', autospec=True)
@mock.patch('tron.bin.recover_batch.read_last_yaml_entries', autospec=True)
@pytest.mark.parametrize('line,exit_code,is_running,error_msg', [
    (  # action runner finishes successfully
        {'return_code': 0, 'runner_pid': 12345},
        0,
        False,
        None
    ), (  # action runner is killed
        {'return_code': -9, 'runner_pid': 12345},
        9,
        False,
        'Action run killed by signal SIGKILL',
    ), (  # No return code but action_runner pid is not running
        {'runner_pid': 12345},
        1,
        False,
        'Action runner pid 12345 no longer running. Assuming an exit of 1.'
    ), (  # No return code but action_runner pid is running
        {'runner_pid': 12345},
        None,
        True,
        None,
    ), (  # No return code or PID from the file
        {},
        None,
        Exception,
        None,
    )
])
def test_get_exit_code(mock_read_last_yaml_entries, mock_pid_running, line, exit_code, is_running, error_msg):
    fake_path = '/file/path'
    mock_read_last_yaml_entries.return_value = line
    mock_pid_running.side_effect = [is_running]

    actual_exit_code, actual_error_msg = recover_batch.get_exit_code(fake_path)
    assert actual_exit_code == exit_code
    assert actual_error_msg == error_msg


def test_read_last_yaml_roundtrip(mock_file):
    """Check that read_last_yaml_entries returns the same thing that the action
    runner wrote."""
    status = StatusFile(mock_file)
    expected_content = [
        {'return_code': None, 'pid': 10345, 'command': 'foo'},
        {'return_code': 1, 'pid': 10345, 'command': 'foo'},
    ]
    with mock.patch.object(status, 'get_content', side_effect=expected_content):
        with status.wrap(command='echo hello', run_id='job.1.action', proc=mock.Mock()):
            # In the context manager, we've written the first value of get_content.
            first = recover_batch.read_last_yaml_entries(mock_file)
            assert first == expected_content[0]

    # After, we write another status entry. We should return the latest.
    second = recover_batch.read_last_yaml_entries(mock_file)
    assert second == expected_content[1]


@mock.patch.object(recover_batch, 'reactor')
@mock.patch('tron.bin.recover_batch.Queue', autospec=True)
@mock.patch('tron.bin.recover_batch.get_exit_code', autospec=True, return_value=(None, None))
@mock.patch('tron.bin.recover_batch.StatusFileWatcher', autospec=True)
@pytest.mark.parametrize('existing_code,watcher_code', [(None, 1), (123, None)])
def test_run(mock_watcher, mock_get_exit_code, mock_queue, mock_reactor, existing_code, watcher_code):
    mock_get_exit_code.return_value = (existing_code, '')
    mock_queue.return_value.get.return_value = (watcher_code, '')
    mock_path = mock.Mock()
    if existing_code is not None:
        expected = existing_code
    else:
        expected = watcher_code

    with pytest.raises(SystemExit) as e:
        recover_batch.run(mock_path)
        assert e.code == expected

    assert mock_get_exit_code.call_args_list == [mock.call(mock_path)]
    if existing_code is not None:
        assert mock_watcher.call_count == 0
    else:
        assert mock_watcher.call_count == 1



from __future__ import absolute_import
from __future__ import unicode_literals

import mock
from mock import Mock

from testifycompat import setup
from testifycompat import TestCase
from tron.actioncommand import NoActionRunnerFactory
from tron.actioncommand import SubprocessActionRunnerFactory
from tron.core.actionrun import ActionRun
from tron.core.actionrun import MesosActionRun
from tron.core.actionrun import SSHActionRun
from tron.core.recovery import filter_action_runs_needing_recovery
from tron.core.recovery import launch_recovery_actionruns_for_job_runs
from tron.utils import timeutils


class TestRecovery(TestCase):
    @setup
    def fake_action_runs(self):
        mock_unknown_machine = Mock(autospec=True)
        mock_ok_machine = Mock(autospec=True)

        mock_unknown_machine.state = ActionRun.UNKNOWN
        mock_ok_machine.state = ActionRun.SUCCEEDED
        self.action_runs = [
            SSHActionRun(
                job_run_id="test.unknown",
                name="test.unknown",
                node=Mock(),
                machine=mock_unknown_machine,
                end_time=timeutils.current_time(),
            ),
            SSHActionRun(
                job_run_id="test.succeeded",
                name="test.succeeded",
                node=Mock(),
                machine=mock_ok_machine,
            ),
            MesosActionRun(
                job_run_id="test.succeeded",
                name="test.succeeded",
                node=Mock(),
                machine=mock_ok_machine,
            ),
            MesosActionRun(
                job_run_id="test.unknown-mesos",
                name="test.unknown-mesos",
                node=Mock(),
                machine=mock_unknown_machine,
            ),
            MesosActionRun(
                job_run_id="test.unknown-mesos-done",
                name="test.unknown-mesos-done",
                node=Mock(),
                machine=mock_unknown_machine,
                end_time=timeutils.current_time(),
            ),
        ]

    def test_filter_action_runs_needing_recovery(self):
        assert filter_action_runs_needing_recovery(self.action_runs) == (
            [self.action_runs[0]],
            [self.action_runs[3]],
        )

    @mock.patch('tron.core.recovery.filter_action_runs_needing_recovery', autospec=True)
    def test_launch_recovery_actionruns_for_job_runs(self, mock_filter):
        mock_actions = (
            [
                mock.Mock(
                    action_runner=NoActionRunnerFactory(), spec=SSHActionRun
                ),
                mock.Mock(
                    action_runner=SubprocessActionRunnerFactory(
                        status_path='/tmp/foo', exec_path=('/tmp/foo')
                    ),
                    spec=SSHActionRun,
                ),
            ],
            [
                mock.Mock(
                    action_runner=NoActionRunnerFactory(), spec=MesosActionRun
                ),
            ],
        )

        mock_filter.return_value = mock_actions
        mock_action_runner = mock.Mock(autospec=True)

        mock_job_run = mock.Mock()
        launch_recovery_actionruns_for_job_runs([mock_job_run],
                                                mock_action_runner)
        ssh_runs = mock_actions[0]
        for run in ssh_runs:
            assert run.recover.call_count == 1

        mesos_run = mock_actions[1][0]
        assert mesos_run.recover.call_count == 1

    @mock.patch('tron.core.recovery.filter_action_runs_needing_recovery', autospec=True)
    def test_launch_recovery_actionruns_empty_job_run(self, mock_filter):
        """_action_runs=None shouldn't prevent other job runs from being recovered"""
        empty_job_run = mock.Mock(_action_runs=None)
        other_job_run = mock.Mock(_action_runs=[mock.Mock()])
        mock_action_runner = mock.Mock()
        mock_filter.return_value = ([], [])

        launch_recovery_actionruns_for_job_runs(
            [empty_job_run, other_job_run],
            mock_action_runner,
        )
        mock_filter.assert_called_with(other_job_run._action_runs)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import unicode_literals

import datetime

import mock

from testifycompat import assert_equal
from testifycompat import assert_raises
from testifycompat import run
from testifycompat import TestCase
from tron.config import config_utils
from tron.config import ConfigError
from tron.config import schedule_parse


class TestPadSequence(TestCase):
    def test_pad_sequence_short(self):
        expected = [0, 1, 2, 3, None, None]
        assert_equal(schedule_parse.pad_sequence(range(4), 6), expected)

    def test_pad_sequence_long(self):
        expected = [0, 1, 2, 3]
        assert_equal(schedule_parse.pad_sequence(range(6), 4), expected)

    def test_pad_sequence_exact(self):
        expected = [0, 1, 2, 3]
        assert_equal(schedule_parse.pad_sequence(range(4), 4), expected)

    def test_pad_sequence_empty(self):
        expected = ["a", "a"]
        assert_equal(schedule_parse.pad_sequence([], 2, "a"), expected)

    def test_pad_negative_size(self):
        assert_equal(schedule_parse.pad_sequence([], -2, "a"), [])


class TestScheduleConfigFromString(TestCase):
    @mock.patch(
        'tron.config.schedule_parse.parse_groc_expression',
        autospec=True,
    )
    def test_groc_config(self, mock_parse_groc):
        schedule = 'every Mon,Wed at 12:00'
        context = config_utils.NullConfigContext
        config = schedule_parse.schedule_config_from_string(schedule, context)
        assert_equal(config, mock_parse_groc.return_value)
        generic_config = schedule_parse.ConfigGenericSchedule(
            'groc daily',
            schedule,
            None,
        )
        mock_parse_groc.assert_called_with(generic_config, context)


class TestValidScheduler(TestCase):
    @mock.patch('tron.config.schedule_parse.schedulers', autospec=True)
    def assert_validation(self, schedule, expected, mock_schedulers):
        context = config_utils.NullConfigContext
        config = schedule_parse.valid_schedule(schedule, context)
        mock_schedulers.__getitem__.assert_called_with('cron')
        func = mock_schedulers.__getitem__.return_value
        assert_equal(config, func.return_value)
        func.assert_called_with(expected, context)

    def test_cron_from_dict(self):
        schedule = {'type': 'cron', 'value': '* * * * *'}
        config = schedule_parse.ConfigGenericSchedule(
            'cron',
            schedule['value'],
            datetime.timedelta(),
        )
        self.assert_validation(schedule, config)

    def test_cron_from_dict_with_jitter(self):
        schedule = {'type': 'cron', 'value': '* * * * *', 'jitter': '5 min'}
        config = schedule_parse.ConfigGenericSchedule(
            'cron',
            schedule['value'],
            datetime.timedelta(minutes=5),
        )
        self.assert_validation(schedule, config)


class TestValidCronScheduler(TestCase):
    _suites = ['integration']

    def validate(self, line):
        config = schedule_parse.ConfigGenericSchedule('cron', line, None)
        context = config_utils.NullConfigContext
        return schedule_parse.valid_cron_scheduler(config, context)

    def test_valid_config(self):
        config = self.validate('5 0 L * *')
        assert_equal(config.minutes, [5])
        assert_equal(config.months, None)
        assert_equal(config.monthdays, ['LAST'])

    def test_invalid_config(self):
        assert_raises(ConfigError, self.validate, '* * *')


class TestValidDailyScheduler(TestCase):
    def validate(self, config):
        context = config_utils.NullConfigContext
        config = schedule_parse.ConfigGenericSchedule('daily', config, None)
        return schedule_parse.valid_daily_scheduler(config, context)

    def assert_parse(self, config, expected):
        config = self.validate(config)
        expected = schedule_parse.ConfigDailyScheduler(*expected, jitter=None)
        assert_equal(config, expected)

    def test_valid_daily_scheduler_start_time(self):
        expected = ('14:32 ', 14, 32, 0, set())
        self.assert_parse('14:32', expected)

    def test_valid_daily_scheduler_just_days(self):
        expected = ("00:00:00 MWS", 0, 0, 0, {1, 3, 6})
        self.assert_parse("00:00:00 MWS", expected)

    def test_valid_daily_scheduler_time_and_day(self):
        expected = ("17:02:44 SU", 17, 2, 44, {0, 6})
        self.assert_parse("17:02:44 SU", expected)

    def test_valid_daily_scheduler_invalid_start_time(self):
        assert_raises(ConfigError, self.validate, "5 MWF")
        assert_raises(ConfigError, self.validate, "05:30:45:45 MWF")
        assert_raises(ConfigError, self.validate, "25:30:45 MWF")

    def test_valid_daily_scheduler_invalid_days(self):
        assert_raises(ConfigError, self.validate, "SUG")
        assert_raises(ConfigError, self.validate, "3")


if __name__ == "__main__":
    run()



#!/usr/bin/env python3.6
import argparse
import logging
import os
import sys
import time

import pytz

from tron.config import manager
from tron.config import schema
from tron.serialize.runstate.dynamodb_state_store import DynamoDBStateStore

# Default values for arguments
DEFAULT_WORKING_DIR = '/var/lib/tron/'
DEFAULT_CONF_PATH = 'config/'
DEFAULT_STALENESS_THRESHOLD = 1800
log = logging.getLogger('check_tron_datastore_staleness')


def get_last_run_time(job):
    '''
    Get all sorted timestamps, and only count the actions that actually ran
    '''
    timestamps = []
    job_runs = job['runs']
    for run in job_runs:
        for action in run['runs']:
            if action.get('start_time') and action.get('state') != 'scheduled':
                timestamps.append(action.get('start_time'))
    return max(timestamps) if timestamps else None


def parse_cli():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-w",
        "--working-dir",
        default=DEFAULT_WORKING_DIR,
        help="Working directory for the Tron daemon, default %(default)s",
    )
    parser.add_argument(
        "-c",
        "--config-path",
        default=DEFAULT_CONF_PATH,
        help="File path to the Tron configuration file",
    )
    parser.add_argument(
        "--job-name",
        required=True,
        help="The job name to read timestamp from",
    )
    parser.add_argument(
        "--staleness-threshold",
        default=DEFAULT_STALENESS_THRESHOLD,
        help="how long (in seconds) to wait to alert after the last timestamp of the job",
    )
    args = parser.parse_args()
    args.working_dir = os.path.abspath(args.working_dir)
    args.config_path = os.path.join(
        args.working_dir,
        args.config_path,
    )
    return args


def read_config(args):
    return manager.ConfigManager(args.config_path).load().get_master().state_persistence


def main():
    # Fetch configs. You can find the arguments in puppet.
    args = parse_cli()
    persistence_config = read_config(args)
    store_type = schema.StatePersistenceTypes(persistence_config.store_type)
    job_name = args.job_name

    # Alert for DynamoDB
    if store_type == schema.StatePersistenceTypes.dynamodb:
        # Fetch job state from dynamodb
        dynamodb_region = persistence_config.dynamodb_region
        table_name = persistence_config.table_name
        store = DynamoDBStateStore(table_name, dynamodb_region)
        key = store.build_key('job_state', job_name)
        try:
            job = store.restore([key])[key]
        except Exception as e:
            logging.exception(f'UNKN: Failed to retreive status for job {job_name} due to {e}')
            sys.exit(3)

        # Exit if the job never runs.
        last_run_time = get_last_run_time(job)
        if not last_run_time:
            logging.error(f'WARN: No last run for {key} found. If the job was just added, it might take some time for it to run')
            sys.exit(1)

        # Alert if timestamp is not updated after staleness_threshold
        stateless_for_secs = time.time() - last_run_time.astimezone(pytz.utc).timestamp()
        if stateless_for_secs > args.staleness_threshold:
            logging.error(f'CRIT: {key} has not been updated in DynamoDB for {stateless_for_secs} seconds')
            sys.exit(2)
        else:
            logging.info(f"OK: DynamoDB is up to date. It's last updated at {last_run_time}")
            sys.exit(0)
    # Alert for BerkeleyDB
    elif store_type == schema.StatePersistenceTypes.shelve:
        os.execl('/usr/lib/nagios/plugins/check_file_age', '/nail/tron/tron_state', '-w', str(args.staleness_threshold), '-c', str(args.staleness_threshold))
    else:
        logging.exception(f'UNKN: Not designed to check this type of datastore: {store_type}')
        sys.exit(3)


if __name__ == '__main__':
    main()



import pickle
from unittest import mock

import boto3
import pytest
from moto import mock_dynamodb2
from moto.dynamodb2.responses import dynamo_json_dump

from testifycompat import assert_equal
from tron.serialize.runstate.dynamodb_state_store import DynamoDBStateStore


def mock_transact_write_items(self):
    """
    This mocks moto.dynamodb2.responses.DynamoHandler.transact_write_items,
    which is used to mock dynamodb client. This function calls put_item,
    update_item, and delete_item based on the arguments of transact_write_item.
    """

    def put_item(item):
        name = item['TableName']
        record = item['Item']
        return self.dynamodb_backend.put_item(name, record)

    def delete_item(item):
        name = item['TableName']
        keys = item['Key']
        return self.dynamodb_backend.delete_item(name, keys)

    def update_item(item):
        name = item['TableName']
        key = item['Key']
        update_expression = item.get('UpdateExpression')
        attribute_updates = item.get('AttributeUpdates')
        expression_attribute_names = item.get('ExpressionAttributeNames', {})
        expression_attribute_values = item.get('ExpressionAttributeValues', {})
        return self.dynamodb_backend.update_item(
            name, key, update_expression, attribute_updates, expression_attribute_names,
            expression_attribute_values)

    transact_items = self.body['TransactItems']

    for transact_item in transact_items:
        if 'Put' in transact_item:
            put_item(transact_item['Put'])
        elif 'Update' in transact_item:
            update_item(transact_item['Update'])
        elif 'Delete' in transact_item:
            delete_item(transact_item['Delete'])

    return dynamo_json_dump({})


@pytest.fixture(autouse=True)
def store():
    with mock.patch(
        'moto.dynamodb2.responses.DynamoHandler.transact_write_items',
        new=mock_transact_write_items, create=True
    ), mock_dynamodb2():
        dynamodb = boto3.resource('dynamodb', region_name='us-west-2')
        table_name = 'tmp'
        store = DynamoDBStateStore(table_name, 'us-west-2', stopping=True)
        store.table = dynamodb.create_table(
            TableName=table_name,
            KeySchema=[
                {
                    'AttributeName': 'key',
                    'KeyType': 'HASH'  # Partition key
                },
                {
                    'AttributeName': 'index',
                    'KeyType': 'RANGE'  # Sort key
                }
            ],
            AttributeDefinitions=[
                {
                    'AttributeName': 'key',
                    'AttributeType': 'S'
                },
                {
                    'AttributeName': 'index',
                    'AttributeType': 'N'
                },

            ],
            ProvisionedThroughput={
                'ReadCapacityUnits': 10,
                'WriteCapacityUnits': 10
            }
        )
        store.client = boto3.client('dynamodb', region_name='us-west-2')
        # Has to be yield here for moto to work
        yield store


@pytest.fixture
def small_object():
    yield pickle.dumps({'this': 'data'})


@pytest.fixture
def large_object():
    yield pickle.dumps([i for i in range(1000000)])


@pytest.mark.usefixtures("store", "small_object", "large_object")
class TestDynamoDBStateStore:
    def test_save(self, store, small_object, large_object):
        key_value_pairs = [
            (
                store.build_key("DynamoDBTest", "two"),
                small_object,
            ),
            (
                store.build_key("DynamoDBTest2", "four"),
                small_object,
            ),
        ]
        store.save(key_value_pairs)
        store._consume_save_queue()

        keys = [store.build_key("DynamoDBTest", "two"), store.build_key("DynamoDBTest2", "four")]
        vals = store.restore(keys)
        for key, value in key_value_pairs:
            assert_equal(vals[key], value)

    def test_save_more_than_4KB(self, store, small_object, large_object):
        key_value_pairs = [
            (
                store.build_key("DynamoDBTest", "two"),
                large_object
            )
        ]
        store.save(key_value_pairs)
        store._consume_save_queue()

        keys = [store.build_key("DynamoDBTest", "two")]
        vals = store.restore(keys)
        for key, value in key_value_pairs:
            assert_equal(vals[key], value)

    def test_restore_more_than_4KB(self, store, small_object, large_object):
        keys = [store.build_key("thing", i) for i in range(3)]
        value = pickle.loads(large_object)
        pairs = zip(keys, (value for i in range(len(keys))))
        store.save(pairs)
        store._consume_save_queue()

        vals = store.restore(keys)
        for key in keys:
            assert_equal(pickle.dumps(vals[key]), large_object)

    def test_restore(self, store, small_object, large_object):
        keys = [store.build_key("thing", i) for i in range(3)]
        value = pickle.loads(small_object)
        pairs = zip(keys, (value for i in range(len(keys))))
        store.save(pairs)
        store._consume_save_queue()

        vals = store.restore(keys)
        for key in keys:
            assert_equal(pickle.dumps(vals[key]), small_object)

    def test_delete(self, store, small_object, large_object):
        keys = [store.build_key("thing", i) for i in range(3)]
        value = pickle.loads(large_object)
        pairs = zip(keys, (value for i in range(len(keys))))
        store.save(pairs)

        for key in pairs:
            store._delete_item(key)

        for key in pairs:
            assert_equal(store._get_num_of_partitions(key), 0)

    def test_retry_saving(self, store, small_object, large_object):
        with mock.patch(
            'moto.dynamodb2.responses.DynamoHandler.transact_write_items',
            side_effect=KeyError('foo')
        ) as mock_failed_write:
            keys = [store.build_key("thing", i) for i in range(1)]
            value = pickle.loads(small_object)
            pairs = zip(keys, (value for i in range(len(keys))))
            try:
                store.save(pairs)
            except Exception:
                assert_equal(mock_failed_write.call_count, 3)

    def test_retry_reading(self, store, small_object, large_object):
        unprocessed_value = {
            'Responses': {
                store.name: [
                    {
                        'index': {'N': '0'},
                        'key': {'S': 'thing 0'}
                    }
                ]
            },
            'UnprocessedKeys':
            {
                store.name: {
                    'ConsistentRead': True,
                    'Keys': [{
                        'index': {'N': '0'},
                        'key': {'S': 'thing 0'}
                    }]
                }
            },
            'ResponseMetadata': {}
        }
        keys = [store.build_key("thing", i) for i in range(1)]
        value = pickle.loads(small_object)
        pairs = zip(keys, (value for i in range(len(keys))))
        store.save(pairs)
        with mock.patch.object(
            store.client,
            'batch_get_item',
            return_value=unprocessed_value
        ) as mock_failed_read:
            try:
                store.restore(keys)
            except Exception:
                assert_equal(mock_failed_read.call_count, 11)



"""
Compatibility fixtures for migrating code from testify to py.test

.. note::

    ``class_`` fixtures must be applied to @classmethods. py.test will not run
    a class_* fixture that is not attached to a class-level method, so your
    tests will probably fail.
"""
import pytest


def setup(func):
    return pytest.fixture(autouse=True)(func)


def setup_teardown(func):
    return pytest.yield_fixture(autouse=True)(func)


def teardown(func):
    def teardown_(*args, **kwargs):
        yield
        func(*args, **kwargs)
    return pytest.yield_fixture(autouse=True)(teardown_)


def class_setup(func):
    return pytest.fixture(autouse=True, scope='class')(func)


def class_setup_teardown(func):
    return pytest.yield_fixture(autouse=True, scope='class')(func)


def class_teardown(func):
    def teardown_(*args, **kwargs):
        yield
        func(*args, **kwargs)
    return pytest.yield_fixture(autouse=True, scope='class')(teardown_)


def suite(name, reason=None):
    """Translate a :func:`testify.suite` decorator into the appropriate
    :mod:`pytest.mark` call. For the disabled suite this results in a
    skipped test. For other suites it will return a  `pytest.mark.<name>`
    decorator.
    """
    if name == 'disabled':
        return pytest.mark.skipif(True, reason=reason)

    return getattr(pytest.mark, name)



"""
Web Controllers for the API.
"""
import logging

from tron import yaml
from tron.eventbus import EventBus

log = logging.getLogger(__name__)


class UnknownCommandError(Exception):
    """Exception raised when a controller received an unknown command."""


class JobCollectionController(object):
    def __init__(self, job_collection):
        self.job_collection = job_collection

    def handle_command(self, command, old_name=None, new_name=None):
        if command == 'move':
            if old_name not in self.job_collection.get_names():
                return f"Error: {old_name} doesn't exist"
            if new_name in self.job_collection.get_names():
                return f"Error: {new_name} exists already"
            return self.job_collection.move(old_name, new_name)

        raise UnknownCommandError(f"Unknown command {command}. Try running this on an individual job or action run id")


class ActionRunController(object):

    mapped_commands = {
        'start',
        'success',
        'cancel',
        'fail',
        'skip',
        'stop',
        'kill',
        'retry',
        'recover',
    }

    def __init__(self, action_run, job_run):
        self.action_run = action_run
        self.job_run = job_run

    def handle_command(self, command):
        if command not in self.mapped_commands:
            raise UnknownCommandError(f"Unknown command {command}. You can only do one of the following to Action runs: {self.mapped_commands}")

        if command == 'start' and self.job_run.is_scheduled:
            return (
                "Action run cannot be started if its job run is still "
                "scheduled."
            )

        if command == 'recover' and not self.action_run.is_unknown:
            return (
                "Action run cannot be recovered if its state is not unknown."
            )

        if command in ('stop', 'kill'):
            return self.handle_termination(command)

        if command == 'retry':
            return self.handle_retry()

        if getattr(self.action_run, command)():
            msg = "%s now in state %s"
            return msg % (self.action_run, self.action_run.state)

        msg = "Failed to %s on %s. State is %s."
        return msg % (command, self.action_run, self.action_run.state)

    def handle_termination(self, command):
        try:
            # Extra message is only used for killing mesos action as warning so far.
            extra_msg = getattr(self.action_run, command)()
            msg = "Attempting to %s %s"
            if extra_msg is not None:
                msg = msg + "\n" + extra_msg
            return msg % (command, self.action_run)
        except NotImplementedError as e:
            msg = "Failed to %s: %s"
            return msg % (command, e)

    def handle_retry(self):
        cleanup_run = self.job_run.action_runs.cleanup_action_run
        if cleanup_run and cleanup_run.is_done:
            return "JobRun has run a cleanup action, use rerun instead"

        if self.action_run.retry():
            return "Retrying %s" % self.action_run
        else:
            return "Failed to schedule retry for %s" % self.action_run


class JobRunController(object):

    mapped_commands = {'start', 'success', 'cancel', 'fail', 'stop'}

    def __init__(self, job_run, job_scheduler):
        self.job_run = job_run
        self.job_scheduler = job_scheduler

    def handle_command(self, command):
        if command == 'restart' or command == 'rerun':
            runs = self.job_scheduler.manual_start(self.job_run.run_time)
            return "Created %s" % ",".join(str(run) for run in runs)

        if command in self.mapped_commands:
            if getattr(self.job_run, command)():
                return "%s now in state %s" % (
                    self.job_run,
                    self.job_run.state,
                )

            msg = "Failed to %s, %s in state %s"
            return msg % (command, self.job_run, self.job_run.state)

        if command == "retry":
            raise UnknownCommandError("Error: Job runs cannot be retried, only individual actions can. Did you mean 'rerun'?")
        else:
            raise UnknownCommandError(f"Unknown command {command}. Only one of the following applies to a Job run: {self.mapped_commands}")


class JobController(object):
    def __init__(self, job_scheduler):
        self.job_scheduler = job_scheduler

    def handle_command(self, command, run_time=None):
        if command == 'enable':
            self.job_scheduler.enable()
            return "%s is enabled" % self.job_scheduler.get_job()

        elif command == 'disable':
            self.job_scheduler.disable()
            return "%s is disabled" % self.job_scheduler.get_job()

        elif command == 'start':
            runs = self.job_scheduler.manual_start(run_time=run_time)
            return "Created %s" % ",".join(str(run) for run in runs)

        if command == "retry":
            raise UnknownCommandError("Error: A whole Job cannot be retried, only individual actions for a specific job run id can.")
        elif command in ["stop", "success", "cancel", "fail", "stop"]:
            raise UnknownCommandError(f"Error: {command} doesn't apply to a whole Job. Please run this on an individual job run id. Hint: try '{self.job_scheduler.get_job()}.-1' for the latest job id")
        else:
            raise UnknownCommandError(f"Unknown command {command}. Does it apply to a whole job? Try a specific Job id or individual action")


class ConfigController(object):
    """Control config. Return config contents and accept updated configuration
    from the API.
    """

    DEFAULT_NAMED_CONFIG = "\njobs:\n"

    def __init__(self, mcp):
        self.mcp = mcp
        self.config_manager = mcp.get_config_manager()

    def _get_config_content(self, name):
        if name not in self.config_manager:
            return self.DEFAULT_NAMED_CONFIG
        return self.config_manager.read_raw_config(name)

    def read_config(self, name):
        config_content = self._get_config_content(name)
        config_hash = self.config_manager.get_hash(name)
        return dict(config=config_content, hash=config_hash)

    def check_config(self, name, content, config_hash):
        """Update a configuration fragment and reload the MCP."""
        if self.config_manager.get_hash(name) != config_hash:
            return "Configuration update will fail: config is stale, try again"

        try:
            content = yaml.load(content)
            self.config_manager.validate_with_fragment(name, content)
        except Exception as e:
            return "Configuration update will fail: %s" % str(e)

    def update_config(self, name, content, config_hash):
        """Update a configuration fragment and reload the MCP."""
        if self.config_manager.get_hash(name) != config_hash:
            return "Configuration has changed. Please try again."

        old_config = self.read_config(name)['config']
        try:
            self.config_manager.write_config(name, content)
            self.mcp.reconfigure()
        except Exception as e:
            log.error("Configuration update failed: %s" % e)
            log.error("Reconfiguring with the previous good configuration")
            try:
                self.config_manager.write_config(name, old_config)
                self.mcp.reconfigure()
            except Exception as e:
                log.error("Could not restore old config: %s" % e)
                return str(e)
            return str(e)

    def delete_config(self, name, content, config_hash):
        """Delete a configuration fragment and reload the MCP."""
        if self.config_manager.get_hash(name) != config_hash:
            return "Configuration has changed. Please try again."

        if content != "":
            return "Configuration content is not empty, will not delete."

        try:
            self.config_manager.delete_config(name)
            self.mcp.reconfigure()
        except Exception as e:
            log.error("Deleting configuration for %s failed: %s" % (name, e))
            return str(e)

    def get_namespaces(self):
        return self.config_manager.get_namespaces()


class EventsController:
    COMMANDS = {'publish', 'discard'}

    def publish(self, event):
        if not EventBus.instance:
            return dict(error='EventBus disabled')

        if EventBus.has_event(event):
            msg = f"event {event} already published"
            log.warning(msg)
            return dict(response=msg)

        if not EventBus.publish(event):
            msg = f'could not publish {event}'
            log.error(msg)
            return dict(error=msg)

        return dict(response='OK')

    def discard(self, event):
        if not EventBus.instance:
            return dict(error='EventBus disabled')

        if not EventBus.discard(event):
            msg = f"could not discard {event}"
            log.error(msg)
            return dict(error=msg)

        return dict(response='OK')

    def info(self):
        if not EventBus.instance:
            return dict(error='EventBus disabled')

        return dict(response=EventBus.instance.event_log)



"""
 Classes to manage job runs.
"""
import json
import logging
import time
from collections import deque

import tron.metrics as metrics
from tron import command_context
from tron import node
from tron.core.actionrun import ActionRun
from tron.core.actionrun import ActionRunFactory
from tron.serialize import filehandler
from tron.utils import maybe_decode
from tron.utils import next_or_none
from tron.utils import proxy
from tron.utils import timeutils
from tron.utils.observer import Observable
from tron.utils.observer import Observer

log = logging.getLogger(__name__)
state_logger = logging.getLogger(f'{__name__}.state_changes')


class Error(Exception):
    pass


class JobRun(Observable, Observer):
    """A JobRun is an execution of a Job.  It has a list of ActionRuns and is
    responsible for starting ActionRuns in the correct order and managing their
    dependencies.
    """

    NOTIFY_DONE = 'notify_done'
    NOTIFY_STATE_CHANGED = 'notify_state_changed'

    context_class = command_context.JobRunContext

    # TODO: use config object
    def __init__(
        self,
        job_name,
        run_num,
        run_time,
        node,
        output_path=None,
        base_context=None,
        action_runs=None,
        action_graph=None,
        manual=None,
    ):
        super(JobRun, self).__init__()
        self.job_name = maybe_decode(job_name)
        self.run_num = run_num
        self.run_time = run_time
        self.node = node
        self.output_path = output_path or filehandler.OutputPath()
        self.output_path.append(str(self.run_num))
        self.action_runs_proxy = None
        self._action_runs = None
        self.action_graph = action_graph
        self.manual = manual

        if action_runs:
            self.action_runs = action_runs

        self.context = command_context.build_context(self, base_context)

    @property
    def id(self):
        return '%s.%s' % (self.job_name, self.run_num)

    @classmethod
    def for_job(cls, job, run_num, run_time, node, manual):
        """Create a JobRun for a job."""
        run = cls(
            job.get_name(),
            run_num,
            run_time,
            node,
            output_path=job.output_path.clone(),
            base_context=job.context,
            action_graph=job.action_graph,
            manual=manual,
        )
        action_runs = ActionRunFactory.build_action_run_collection(
            run,
            job.action_runner,
        )
        run.action_runs = action_runs
        return run

    @classmethod
    def from_state(
        cls,
        state_data,
        action_graph,
        output_path,
        context,
        run_node,
    ):
        """Restore a JobRun from a serialized state."""
        pool_repo = node.NodePoolRepository.get_instance()
        run_node = pool_repo.get_node(state_data.get('node_name'), run_node)
        job_name = state_data['job_name']

        job_run = cls(
            job_name,
            state_data['run_num'],
            state_data['run_time'],
            run_node,
            action_graph=action_graph,
            manual=state_data.get('manual', False),
            output_path=output_path,
            base_context=context,
        )
        action_runs = ActionRunFactory.action_run_collection_from_state(
            job_run,
            state_data['runs'],
            state_data['cleanup_run'],
        )
        job_run.action_runs = action_runs
        return job_run

    @property
    def state_data(self):
        """This data is used to serialize the state of this job run."""
        return {
            'job_name': self.job_name,
            'run_num': self.run_num,
            'run_time': self.run_time,
            'node_name': self.node.get_name() if self.node else None,
            'runs': self.action_runs.state_data,
            'cleanup_run': self.action_runs.cleanup_action_state_data,
            'manual': self.manual,
        }

    def _get_action_runs(self):
        return self._action_runs

    def _set_action_runs(self, run_collection):
        """Store action runs and register callbacks."""
        if self._action_runs is not None:
            raise ValueError("ActionRunCollection already set on %s" % self)

        self._action_runs = run_collection
        for action_run in run_collection.action_runs_with_cleanup:
            self.watch(action_run)
            action_run.setup_subscriptions()

        self.action_runs_proxy = proxy.AttributeProxy(
            run_collection,
            [
                'queue',
                'cancel',
                'success',
                'fail',
                'start_time',
                'end_time',
            ],
        )

    def _del_action_runs(self):
        self._action_runs = None
        self.action_runs_proxy = None

    action_runs = property(
        _get_action_runs,
        _set_action_runs,
        _del_action_runs,
    )

    def seconds_until_run_time(self):
        run_time = self.run_time
        if run_time.tzinfo:
            now = timeutils.current_time(tz=run_time.tzinfo)
        else:
            now = timeutils.current_time()
        return max(0, timeutils.delta_total_seconds(run_time - now))

    def start(self):
        self.log_state_update(state='start')
        if self._do_start():
            return True

    def _do_start(self):
        log.info(f"{self} starting")
        self.action_runs.ready()
        if any(self._start_action_runs()):
            log.info(f'{self} started')
            return True

    def stop(self):
        if self.action_runs.is_done:
            return
        self.action_runs.stop()

    def _start_action_runs(self):
        """Start all startable action runs, and return any that were
        successfully started.
        """
        started_runs = [
            action_run
            for action_run in self.action_runs.get_startable_action_runs()
            if action_run.start()
        ]

        if not started_runs:
            for r in self.action_runs:
                if r.is_blocked_on_trigger:
                    log.debug(f"{r} is blocked on triggers: {r.remaining_triggers}")

        return started_runs

    def handle_action_run_state_change(self, action_run: ActionRun, event):
        """Handle events triggered by JobRuns."""
        log.info(f"{self} got an event: {event}")
        metrics.meter(f'tron.actionrun.{event}')

        if event == ActionRun.NOTIFY_TRIGGER_READY:
            if self.is_scheduled or self.is_queued:
                log.info(f"{self} triggers are satisfied but run not started yet")
                return

            started = self._start_action_runs()
            if any(started):
                log.info(
                    f"{self} action runs triggered: "
                    f"{', '.join(str(s) for s in started)}"
                )
            return

        # propagate all state changes (from action runs) up to state serializer
        self.notify(self.NOTIFY_STATE_CHANGED)
        self.log_state_update(
            state=action_run.state,
            action_name=action_run.name,
        )

        if not action_run.is_done:
            return

        if action_run.is_skipped and self.action_runs.is_scheduled:
            return

        if not action_run.is_broken:
            started = self._start_action_runs()
            if any(started):
                log.info(
                    f"{self} action runs started: "
                    f"{', '.join(str(s) for s in started)}"
                )
                return

        if not self.action_runs.is_done:
            log.info(f"{self} still has running or waiting actions")
            return

        # If we can't make any progress, we're done
        cleanup_run: ActionRun = self.action_runs.cleanup_action_run
        if not cleanup_run or cleanup_run.is_done:
            return self.finalize()

        cleanup_run.start()

    handler = handle_action_run_state_change

    def finalize(self):
        """The last step of a JobRun. Called when the cleanup action
        completes or if the job has no cleanup action, called once all action
        runs have reached a 'done' state.

        Triggers an event to notifies the Job that is is done.
        """
        if self.action_runs.is_failed:
            log.error(f'{self} failed')
        else:
            log.info(f'{self} succeeded')

        # Notify Job that this JobRun is complete
        self.notify(self.NOTIFY_DONE)
        self.log_state_update(state=self.state)

    def cleanup(self):
        """Cleanup any resources used by this JobRun."""
        log.info(f'{self} removed')
        self.clear_observers()
        self.action_runs.cleanup()
        self.node = None
        self.action_graph = None
        self._del_action_runs()
        self.output_path.delete()

    def get_action_run(self, action_name):
        return self.action_runs.get(action_name)

    def log_state_update(self, state, action_name=None):
        if action_name is None:
            state = f'job_{state}'
        else:
            action_name = str(action_name)

        data = {
            'job_name': str(self.job_name),
            'run_num': str(self.run_num),
            'action_name': action_name,
            'state': str(state),
            'timestamp': time.time(),
        }
        state_logger.info(json.dumps(data))

    @property
    def state(self):
        """The overall state of this job run. Based on the state of its actions.
        """
        if not self.action_runs:
            log.info(f"{self} has no action runs to determine state")
            return ActionRun.UNKNOWN

        if self.action_runs.is_complete:
            return ActionRun.SUCCEEDED
        if self.action_runs.is_cancelled:
            return ActionRun.CANCELLED
        if self.action_runs.is_running:
            return ActionRun.RUNNING
        if self.action_runs.is_starting:
            return ActionRun.STARTING
        if self.action_runs.is_failed:
            return ActionRun.FAILED
        if self.action_runs.is_waiting and self.action_runs.is_blocked_on_trigger:
            return ActionRun.WAITING
        if self.action_runs.is_scheduled:
            return ActionRun.SCHEDULED
        if self.action_runs.is_queued:
            return ActionRun.QUEUED

        return ActionRun.UNKNOWN

    def cancel(self):
        return self.action_runs.cancel()

    def __getattr__(self, name):
        if name.startswith('is_'):
            state_name = name[3:]
            if state_name not in ActionRun.STATE_MACHINE.states:
                raise RuntimeError(f"{state_name} not in ActionRun.VALID_STATES")
            return self.state == state_name
        elif self.action_runs_proxy:
            return self.action_runs_proxy.perform(name)
        else:
            raise AttributeError(name)

    def __str__(self):
        return f"JobRun:{self.id}"


class JobRunCollection(object):
    """A JobRunCollection is a deque of JobRun objects. Responsible for
    ordering and logic related to a group of JobRuns which should all be runs
    for the same Job.

    A JobRunCollection is created in two stages. First it's populated from a
    configuration object, and second its state is loaded from a serialized
    state dict.

    Runs in a JobRunCollection should always remain sorted by their run_num.
    """

    def __init__(self, run_limit):
        self.run_limit = run_limit
        self.runs = deque()

    @classmethod
    def from_config(cls, job_config):
        """Factory method for creating a JobRunCollection from a config."""
        return cls(job_config.run_limit)

    def build_new_run(self, job, run_time, node, manual=False):
        """Create a new run for the job, add it to the runs list,
        and return it.
        """
        run_num = self.next_run_num()
        run = JobRun.for_job(job, run_num, run_time, node, manual)
        log.info(f"{run} created on {node.name} at {run_time}")
        self.runs.appendleft(run)
        self.remove_old_runs()
        return run

    def cancel_pending(self):
        """Find any queued or scheduled runs and cancel them."""
        for pending in self.get_pending():
            pending.cancel()

    def remove_pending(self):
        """Remove pending runs from the run list."""
        for pending in list(self.get_pending()):
            pending.cleanup()
            self.runs.remove(pending)

    def get_run_by_state(self, state):
        """Returns the most recent run which matches the state."""
        return next_or_none(r for r in self.runs if r.state == state)

    def get_run_by_num(self, num):
        """Return a the run with run number which matches num."""
        return next_or_none(r for r in self.runs if r.run_num == num)

    def get_run_by_index(self, index):
        """Return the job run at index. Jobs are indexed from oldest to newest.
        """
        try:
            return self.runs[index * -1 - 1]
        except IndexError:
            return None

    def get_newest(self, include_manual=True):
        """Returns the most recently created JobRun."""
        return next_or_none(
            r for r in self.runs if include_manual or not r.manual
        )

    def get_pending(self):
        """Return the job runs that are queued or scheduled."""
        return [r for r in self.runs if r.state in (ActionRun.SCHEDULED, ActionRun.QUEUED)]

    @property
    def has_pending(self):
        return any(self.get_pending())

    def get_active(self, node=None):
        active_states = {ActionRun.RUNNING, ActionRun.STARTING, ActionRun.WAITING}
        return [
            r for r in self.runs if (r.state in active_states) and
            (not node or r.node == node)
        ]

    def get_first_queued(self, node=None):
        return next_or_none(
            r for r in reversed(self.runs)
            if (not node or r.node == node) and r.state == ActionRun.QUEUED
        )

    def get_scheduled(self):
        return [r for r in self.runs if r.state == ActionRun.SCHEDULED]

    def next_run_num(self):
        """Return the next run number to use."""
        if not self.runs:
            return 0
        return max(r.run_num for r in self.runs) + 1

    def remove_old_runs(self):
        """Remove old runs to reduce the number of completed runs
        to within RUN_LIMIT.
        """
        while len(self.runs) > self.run_limit:
            run = self.runs.pop()
            run.cleanup()

    def get_action_runs(self, action_name):
        return [job_run.get_action_run(action_name) for job_run in self.runs]

    @property
    def state_data(self):
        """Return the state data to serialize."""
        return [r.state_data for r in self.runs]

    @property
    def last_success(self):
        return self.get_run_by_state(ActionRun.SUCCEEDED)

    @property
    def next_run(self):
        return self.get_run_by_state(ActionRun.SCHEDULED)

    def __iter__(self):
        return iter(self.runs)

    def __str__(self):
        return "%s[%s]" % (
            type(self).__name__,
            ', '.join("%s(%s)" % (r.run_num, r.state) for r in self.runs),
        )


def job_runs_from_state(
    runs,
    action_graph,
    output_path,
    context,
    node_pool,
):
    return [
        JobRun.from_state(
            run,
            action_graph,
            output_path.clone(),
            context,
            node_pool.next(),
        ) for run in runs
    ]



import queue

from twisted.internet import defer


class PyDeferredQueue(defer.DeferredQueue):
    """
    Implements the stdlib queue.Queue get/put interface with a DeferredQueue.
    """

    def __init__(self, *args, **kwargs):
        super(PyDeferredQueue, self).__init__(*args, **kwargs)

    def put(self, item, block=None, timeout=None):
        # Call from reactor thread so callbacks from get() will be executed
        # on the reactor thread, even if this is called from another thread.
        from twisted.internet import reactor
        try:
            reactor.callFromThread(super(PyDeferredQueue, self).put, item)
        except defer.QueueOverflow:
            raise queue.Full

    def get(self, block=None, timeout=None):
        try:
            return super(PyDeferredQueue, self).get()
        except defer.QueueUnderflow:
            raise queue.Empty



from collections import defaultdict
from collections import namedtuple

from tron.core.action import Action
from tron.core.actiongraph import ActionGraph
from tron.utils import maybe_decode

AdjListEntry = namedtuple('AdjListEntry', ['action_name', 'is_trigger'])


class JobGraph(object):
    """ A JobGraph stores the entire DAG of jobs and actions, including
    cross-job dependencies (aka triggers)
    """

    def __init__(self, config_container, should_validate_missing_dependency=False):
        """ Build an adjacency list and a reverse adjacency list for the graph,
        and store all the actions as well as which actions belong to which job
        """
        self.action_map = {}
        self._actions_for_job = defaultdict(list)
        self._adj_list = defaultdict(list)
        self._rev_adj_list = defaultdict(list)

        all_actions = set()
        for job_name, job_config in config_container.get_jobs().items():
            for action_name, action_config in job_config.actions.items():
                full_name = self._save_action(action_name, job_name, action_config)
                all_actions.add(full_name)

                for required_action in action_config.requires or []:
                    required_action_name = f'{job_name}.{required_action}'
                    self._rev_adj_list[full_name].append(AdjListEntry(required_action_name, False))

                for trigger in action_config.triggered_by or []:
                    trigger_action_name = '.'.join(trigger.split('.')[:3])
                    self._rev_adj_list[full_name].append(AdjListEntry(trigger_action_name, True))

                for parent_action, is_trigger in self._rev_adj_list[full_name]:
                    self._adj_list[parent_action].append(AdjListEntry(full_name, is_trigger))

            cleanup_action_config = job_config.cleanup_action
            if cleanup_action_config:
                self._save_action(cleanup_action_config.name, job_name, cleanup_action_config)

        if should_validate_missing_dependency:
            missing_dependent_actions = defaultdict(list)
            for action_name in self._rev_adj_list:
                for dependent_action_entry in self._rev_adj_list[action_name]:
                    if dependent_action_entry.action_name not in all_actions:
                        missing_dependent_actions[dependent_action_entry.action_name].append(action_name)

            error_messages = []
            for action_name, child_action_names in missing_dependent_actions.items():
                error_messages.append(
                    'Action {0} is dependency of actions:\n{1}'.format(
                        action_name,
                        '\n'.join(
                            ['  - {}'.format(child_action_name) for child_action_name in child_action_names]
                        )
                    )
                )

            if error_messages:
                raise ValueError(
                    (
                        'The following actions are dependencies of other actions but missing:\n'
                        '{0}\n'
                        'Please check if you have deleted/renamed any of them or their containing jobs.'
                    ).format(
                        '\n'.join(error_messages),
                    )
                )

    def get_action_graph_for_job(self, job_name):
        """ Traverse the JobGraph for a specific job to construct an ActionGraph for it """
        job_action_map = {}
        required_actions, required_triggers = defaultdict(set), defaultdict(set)

        for action_name in self._actions_for_job[job_name]:
            # Any actions that belong to _this job_ are not prefixed by the job name
            short_action_name = action_name.split('.')[-1]
            job_action_map[short_action_name] = self.action_map[action_name]
            required_actions[short_action_name] = {
                entry.action_name.split('.')[-1]
                for entry in self._rev_adj_list[action_name]
                if not entry.is_trigger
            }

            # We call this twice to build the complete DAG for the job; the first time
            # we search the forward adjacency list and the second time we search the
            # reverse adjancency list.  This ensures we don't miss any triggers
            required_triggers = self._get_required_triggers(action_name, required_triggers)
            required_triggers = self._get_required_triggers(action_name, required_triggers, search_up=False)
        return ActionGraph(job_action_map, required_actions, required_triggers)

    def _save_action(self, action_name, job_name, config):
        action_name = maybe_decode(action_name)
        full_name = f'{job_name}.{action_name}'
        self.action_map[full_name] = Action.from_config(config)
        self._actions_for_job[job_name].append(full_name)
        return full_name

    def _get_required_triggers(self, action_name, triggers, search_up=True):
        stack = [action_name]
        visited = set()

        # Do DFS to search the adjacency list and find all of the required triggers
        # for a particular action
        while stack:
            current_action = stack.pop()
            visited.add(current_action)
            adj_list = self._rev_adj_list if search_up else self._adj_list
            for next_action, is_trigger in adj_list[current_action]:
                if not is_trigger:
                    continue

                if next_action not in visited:
                    stack.append(next_action)

                if current_action == action_name:
                    current_action = current_action.split('.')[-1]

                if search_up:
                    triggers[current_action].add(next_action)
                else:
                    triggers[next_action].add(current_action)

        return triggers



import logging
import time

from tron import actioncommand
from tron import command_context
from tron import node
from tron.config import manager
from tron.core.job import Job
from tron.core.job_collection import JobCollection
from tron.core.job_scheduler import JobSchedulerFactory
from tron.core.jobgraph import JobGraph
from tron.eventbus import EventBus
from tron.mesos import MesosClusterRepository
from tron.serialize.runstate import statemanager

log = logging.getLogger(__name__)


def apply_master_configuration(mapping, master_config):
    def get_config_value(seq):
        return [getattr(master_config, item) for item in seq]

    for entry in mapping:
        func, args = entry[0], get_config_value(entry[1:])
        func(*args)


class MasterControlProgram(object):
    """Central state object for the Tron daemon."""

    def __init__(self, working_dir, config_path):
        super(MasterControlProgram, self).__init__()
        self.jobs = JobCollection()
        self.working_dir = working_dir
        self.config = manager.ConfigManager(config_path)
        self.context = command_context.CommandContext()
        self.state_watcher = statemanager.StateChangeWatcher()
        self.boot_time = time.time()
        log.info('initialized')

    def shutdown(self):
        EventBus.shutdown()
        self.state_watcher.shutdown()

    def reconfigure(self):
        """Reconfigure MCP while Tron is already running."""
        log.info("reconfigured")
        try:
            self._load_config(reconfigure=True)
        except Exception as e:
            log.exception(f"reconfigure failure: {e.__class__.__name__}: {e}")
            raise e

    def _load_config(self, reconfigure=False):
        """Read config data and apply it."""
        with self.state_watcher.disabled():
            self.apply_config(self.config.load(), reconfigure=reconfigure)

    def initial_setup(self):
        """When the MCP is initialized the config is applied before the state.
        In this case jobs shouldn't be scheduled until the state is applied.
        """
        self._load_config()
        self.restore_state(
            actioncommand.create_action_runner_factory_from_config(
                self.config.load().get_master().action_runner
            )
        )
        # Any job with existing state would have been scheduled already. Jobs
        # without any state will be scheduled here.
        self.jobs.run_queue_schedule()

    def apply_config(self, config_container, reconfigure=False):
        """Apply a configuration."""
        master_config_directives = [
            (self.update_state_watcher_config, 'state_persistence'),
            (self.set_context_base, 'command_context'),
            (
                node.NodePoolRepository.update_from_config,
                'nodes',
                'node_pools',
                'ssh_options',
            ),
            (MesosClusterRepository.configure, 'mesos_options'),
            (self.configure_eventbus, 'eventbus_enabled'),
        ]
        master_config = config_container.get_master()
        apply_master_configuration(master_config_directives, master_config)

        self.state_watcher.watch(MesosClusterRepository)

        # TODO: unify NOTIFY_STATE_CHANGE and simplify this
        self.job_graph = JobGraph(config_container)
        factory = self.build_job_scheduler_factory(master_config, self.job_graph)
        self.apply_collection_config(
            config_container.get_jobs(),
            self.jobs,
            Job.NOTIFY_STATE_CHANGE,
            factory,
            reconfigure,
        )

    def apply_collection_config(self, config, collection, notify_type, *args):
        items = collection.load_from_config(config, *args)
        self.state_watcher.watch_all(items, notify_type)

    def build_job_scheduler_factory(self, master_config, job_graph):
        output_stream_dir = master_config.output_stream_dir or self.working_dir
        action_runner = actioncommand.create_action_runner_factory_from_config(
            master_config.action_runner,
        )
        return JobSchedulerFactory(
            self.context,
            output_stream_dir,
            master_config.time_zone,
            action_runner,
            job_graph,
        )

    def update_state_watcher_config(self, state_config):
        """Update the StateChangeWatcher, and save all state if the state config
        changed.
        """
        if self.state_watcher.update_from_config(state_config):
            for job_scheduler in self.jobs:
                self.state_watcher.save_job(job_scheduler.get_job())

    def set_context_base(self, command_context):
        self.context.base = command_context

    def configure_eventbus(self, enabled):
        if enabled:
            if not EventBus.instance:
                EventBus.create(f"{self.working_dir}/_events")
                EventBus.start()
        else:
            EventBus.shutdown()

    def get_job_collection(self):
        return self.jobs

    def get_config_manager(self):
        return self.config

    def restore_state(self, action_runner):
        """Use the state manager to retrieve to persisted state and apply it
        to the configured Jobs.
        """
        log.info('restoring')
        states = self.state_watcher.restore(self.jobs.get_names())
        MesosClusterRepository.restore_state(states.get('mesos_state', {}))

        self.jobs.restore_state(states.get('job_state', {}), action_runner)
        self.state_watcher.save_metadata()

    def __str__(self):
        return "MCP"



import yaml


def dump(*args, **kwargs):
    kwargs['Dumper'] = yaml.CSafeDumper
    return yaml.dump(*args, **kwargs)


def load(*args, **kwargs):
    kwargs['Loader'] = yaml.CSafeLoader
    return yaml.load(*args, **kwargs)


def load_all(*args, **kwargs):
    kwargs['Loader'] = yaml.CSafeLoader
    return yaml.load_all(*args, **kwargs)


safe_dump = dump
safe_load = load
safe_load_all = load_all



"""
 Immutable config schema objects.
"""
from __future__ import absolute_import
from __future__ import unicode_literals

from collections import namedtuple
from enum import Enum

MASTER_NAMESPACE = "MASTER"

CLEANUP_ACTION_NAME = 'cleanup'


def config_object_factory(name, required=None, optional=None):
    """
    Creates a namedtuple which has two additional attributes:
        required_keys:
            all keys required to be set on this configuration object
        optional keys:
            optional keys for this configuration object

    The tuple is created from required + optional
    """
    required = required or []
    optional = optional or []

    config_class = namedtuple(name, required + optional)

    # make last len(optional) args actually optional
    config_class.__new__.__defaults__ = (None, ) * len(optional)
    config_class.required_keys = required
    config_class.optional_keys = optional

    return config_class


TronConfig = config_object_factory(
    name='TronConfig',
    optional=[
        'output_stream_dir',  # str
        'action_runner',  # ConfigActionRunner
        'state_persistence',  # ConfigState
        'command_context',  # dict of str
        'ssh_options',  # ConfigSSHOptions
        'time_zone',  # pytz time zone
        'nodes',  # dict of ConfigNode
        'node_pools',  # dict of ConfigNodePool
        'jobs',  # dict of ConfigJob
        'mesos_options',  # ConfigMesos
        'eventbus_enabled',  # bool or None
    ],
)

NamedTronConfig = config_object_factory(
    name='NamedTronConfig',
    optional=[
        'jobs',  # dict of ConfigJob
    ],
)

ConfigActionRunner = config_object_factory(
    'ConfigActionRunner',
    optional=['runner_type', 'remote_status_path', 'remote_exec_path'],
)

ConfigSSHOptions = config_object_factory(
    name='ConfigSSHOptions',
    optional=[
        'agent',
        'identities',
        'known_hosts_file',
        'connect_timeout',
        'idle_connection_timeout',
        'jitter_min_load',
        'jitter_max_delay',
        'jitter_load_factor',
    ],
)

ConfigNode = config_object_factory(
    name='ConfigNode',
    required=['hostname'],
    optional=['name', 'username', 'port'],
)

ConfigNodePool = config_object_factory('ConfigNodePool', ['nodes'], ['name'])

ConfigState = config_object_factory(
    name='ConfigState',
    required=[
        'name',
        'store_type',
    ],
    optional=[
        'buffer_size',
        'dynamodb_region',
        'table_name',
    ],
)

ConfigMesos = config_object_factory(
    name='ConfigMesos',
    optional=[
        'master_address',
        'master_port',
        'secret_file',
        'principal',
        'role',
        'enabled',
        'default_volumes',
        'dockercfg_location',
        'offer_timeout',
    ],
)

ConfigJob = config_object_factory(
    name='ConfigJob',
    required=[
        'name',  # str
        'node',  # str
        'schedule',  # Config*Scheduler
        'actions',  # dict of ConfigAction
        'namespace',  # str
    ],
    optional=[
        'monitoring',  # dict
        'queueing',  # bool
        'run_limit',  # int
        'all_nodes',  # bool
        'cleanup_action',  # ConfigAction
        'enabled',  # bool
        'allow_overlap',  # bool
        'max_runtime',  # datetime.Timedelta
        'time_zone',  # pytz time zone
        'expected_runtime',  # datetime.Timedelta
    ],
)

ConfigAction = config_object_factory(
    name='ConfigAction',
    required=[
        'name',  # str
        'command',  # str
    ],
    optional=[
        'requires',  # tuple of str
        'node',  # str
        'retries',  # int
        'retries_delay',  # datetime.Timedelta
        'executor',  # str
        'cpus',  # float
        'mem',  # float
        'disk',  # float
        'constraints',  # List of ConfigConstraint
        'docker_image',  # str
        'docker_parameters',  # List of ConfigParameter
        'env',  # dict
        'extra_volumes',  # List of ConfigVolume
        'expected_runtime',  # datetime.Timedelta
        'trigger_downstreams',  # None, bool or dict
        'triggered_by',  # list or None
        'on_upstream_rerun',  # ActionOnRerun or None
        'trigger_timeout',  # datetime.deltatime or None
    ],
)

ConfigCleanupAction = config_object_factory(
    name='ConfigCleanupAction',
    required=[
        'command',  # str
    ],
    optional=[
        'name',  # str
        'node',  # str
        'retries',  # int
        'retries_delay',  # datetime.Timedelta
        'expected_runtime',  # datetime.Timedelta
        'executor',  # str
        'cpus',  # float
        'mem',  # float
        'disk',  # float
        'constraints',  # List of ConfigConstraint
        'docker_image',  # str
        'docker_parameters',  # List of ConfigParameter
        'env',  # dict
        'extra_volumes',  # List of ConfigVolume
        'trigger_downstreams',  # None, bool or dict
        'triggered_by',  # list or None
        'on_upstream_rerun',  # ActionOnRerun or None
        'trigger_timeout',  # datetime.deltatime or None
    ],
)

ConfigConstraint = config_object_factory(
    name='ConfigConstraint',
    required=[
        'attribute',
        'operator',
        'value',
    ],
    optional=[],
)

ConfigVolume = config_object_factory(
    name='ConfigVolume',
    required=[
        'container_path',
        'host_path',
        'mode',
    ],
    optional=[],
)

ConfigParameter = config_object_factory(
    name='ConfigParameter',
    required=[
        'key',
        'value',
    ],
    optional=[],
)


StatePersistenceTypes = Enum(
    'StatePersistenceTypes', dict(shelve='shelve', yaml='yaml', dynamodb='dynamodb')
)

ExecutorTypes = Enum('ExecutorTypes', dict(ssh='ssh', mesos='mesos'))

ActionRunnerTypes = Enum('ActionRunnerTypes', dict(none='none', subprocess='subprocess'))

VolumeModes = Enum('VolumeModes', dict(RO='RO', RW='RW'))

ActionOnRerun = Enum('ActionOnRerun', dict(rerun='rerun'))



from __future__ import absolute_import
from __future__ import unicode_literals

import datetime
import textwrap
from subprocess import CalledProcessError
from textwrap import dedent

import pytest

from testifycompat import assert_equal
from testifycompat import assert_gt
from tests import sandbox
from tron.core import actionrun

BASIC_CONFIG = """
ssh_options:
    agent: true

nodes:
  - name: local
    hostname: 'localhost'

state_persistence:
    name: "state_data.shelve"
    store_type: shelve

"""

SINGLE_ECHO_CONFIG = BASIC_CONFIG + """
jobs:
  - name: "echo_job"
    node: local
    schedule: "cron 0 * * * *"
    actions:
      - name: "echo_action"
        command: "echo 'Echo!'" """

DOUBLE_ECHO_CONFIG = SINGLE_ECHO_CONFIG + """
      - name: "another_echo_action"
        command: "echo 'Today is %(shortdate)s, which is the same
                    as %(year)s-%(month)s-%(day)s' && false" """

ALT_NAMESPACED_ECHO_CONFIG = """
jobs:
  - name: "echo_job"
    node: local
    schedule: "cron 0 * * * *"
    actions:
      - name: "echo_action"
        command: "echo 'Echo!'" """

TOUCH_CLEANUP_FMT = """
    cleanup_action:
      command: "echo 'at last'"
"""


@pytest.mark.skip(reason="We don't have a setup for sandbox tests yet")
class TrondEndToEndTestCase(sandbox.SandboxTestCase):
    def test_end_to_end_basic(self):
        self.start_with_config(SINGLE_ECHO_CONFIG)
        client = self.sandbox.client

        assert_equal(
            self.client.config('MASTER')['config'],
            SINGLE_ECHO_CONFIG,
        )

        # reconfigure and confirm results
        second_config = DOUBLE_ECHO_CONFIG + TOUCH_CLEANUP_FMT
        self.sandbox.tronfig(second_config)
        assert_equal(client.config('MASTER')['config'], second_config)

        # reconfigure, by uploading a third configuration
        self.sandbox.tronfig(ALT_NAMESPACED_ECHO_CONFIG, name='ohce')
        self.sandbox.client.home()

        # run the job and check its output
        echo_job_name = 'MASTER.echo_job'
        job_url = client.get_url(echo_job_name)
        action_url = client.get_url('MASTER.echo_job.1.echo_action')

        self.sandbox.tronctl('start', echo_job_name)

        def wait_on_cleanup():
            return (
                len(client.job(job_url)['runs']) >= 2 and
                client.action_runs(action_url)['state'] ==
                actionrun.ActionRun.SUCCEEDED
            )

        sandbox.wait_on_sandbox(wait_on_cleanup)

        echo_action_run = client.action_runs(action_url)
        another_action_url = client.get_url(
            'MASTER.echo_job.1.another_echo_action',
        )
        other_act_run = client.action_runs(another_action_url)
        assert_equal(
            echo_action_run['state'],
            actionrun.ActionRun.SUCCEEDED,
        )
        assert_equal(echo_action_run['stdout'], ['Echo!'])
        assert_equal(
            other_act_run['state'],
            actionrun.ActionRun.FAILED,
        )

        now = datetime.datetime.now()
        stdout = now.strftime(
            'Today is %Y-%m-%d, which is the same as %Y-%m-%d',
        )
        assert_equal(other_act_run['stdout'], [stdout])

        job_runs_url = client.get_url('%s.1' % echo_job_name)
        assert_equal(
            client.job_runs(job_runs_url)['state'],
            actionrun.ActionRun.FAILED,
        )

    def test_node_reconfig(self):
        job_config = dedent(
            """
            jobs:
                - name: a_job
                  node: local
                  schedule: "cron * * * * *"
                  actions:
                    - name: first_action
                      command: "echo something"
        """
        )
        second_config = dedent(
            """
            ssh_options:
                agent: true

            nodes:
              - name: local
                hostname: '127.0.0.1'

            state_persistence:
                name: "state_data.shelve"
                store_type: shelve

        """
        ) + job_config
        self.start_with_config(BASIC_CONFIG + job_config)

        job_url = self.client.get_url('MASTER.a_job.0')
        sandbox.wait_on_state(
            self.client.job_runs,
            job_url,
            actionrun.ActionRun.SUCCEEDED,
        )

        self.sandbox.tronfig(second_config)

        job_url = self.client.get_url('MASTER.a_job')

        def wait_on_next_run():
            last_run = self.client.job(job_url)['runs'][0]
            return last_run['node']['hostname'] == '127.0.0.1'

        sandbox.wait_on_sandbox(wait_on_next_run)


@pytest.mark.skip(reason="We don't have a setup for sandbox tests yet")
class TronCommandsTestCase(sandbox.SandboxTestCase):
    def test_tronview(self):
        self.start_with_config(SINGLE_ECHO_CONFIG)
        expected = """\nServices:\nNo Services\n\n\nJobs:
            Name       State       Scheduler           Last Success
            MASTER.echo_job   enabled     cron 1:00:00    None
            """

        def remove_line_space(s):
            return [l.replace(' ', '') for l in s.split('\n')]

        actual = self.sandbox.tronview()[0]
        assert_equal(remove_line_space(actual), remove_line_space(expected))

    def test_tronctl_with_job(self):
        self.start_with_config(SINGLE_ECHO_CONFIG + TOUCH_CLEANUP_FMT)
        job_name = 'MASTER.echo_job'
        job_url = self.client.get_url(job_name)
        self.sandbox.tronctl('start', job_name)

        cleanup_url = self.client.get_url('MASTER.echo_job.1.cleanup')
        sandbox.wait_on_state(
            self.client.action_runs,
            cleanup_url,
            actionrun.ActionRun.SUCCEEDED,
        )

        action_run_url = self.client.get_url('MASTER.echo_job.1.echo_action')
        assert_equal(
            self.client.action_runs(action_run_url)['state'],
            actionrun.ActionRun.SUCCEEDED,
        )

        job_run_url = self.client.get_url('MASTER.echo_job.1')
        assert_equal(
            self.client.job_runs(job_run_url)['state'],
            actionrun.ActionRun.SUCCEEDED,
        )

        assert_equal(self.client.job(job_url)['status'], 'enabled')
        self.sandbox.tronctl('disable', job_name)
        sandbox.wait_on_state(self.client.job, job_url, 'disabled', 'status')

    def test_tronfig(self):
        self.start_with_config(SINGLE_ECHO_CONFIG)
        stdout, stderr = self.sandbox.tronfig()
        assert_equal(stdout.rstrip(), SINGLE_ECHO_CONFIG.rstrip())

    def test_tronfig_failure(self):
        self.start_with_config(SINGLE_ECHO_CONFIG)
        bad_config = 'this is not valid: yaml: is it?'

        def test_return_code(exc):
            assert_equal(exc.returncode, 1)

        with pytest.raises(CalledProcessError):
            test_return_code(self.sandbox.tronfig, bad_config)


@pytest.mark.skip(reason="We don't have a setup for sandbox tests yet")
class JobEndToEndTestCase(sandbox.SandboxTestCase):
    def test_cleanup_on_failure(self):
        config = BASIC_CONFIG + dedent(
            """
        jobs:
          - name: "failjob"
            node: local
            schedule: "daily 04:20"
            actions:
              - name: "failaction"
                command: "failplz"
        """
        ) + TOUCH_CLEANUP_FMT
        self.start_with_config(config)

        action_run_url = self.client.get_url('MASTER.failjob.0.failaction')
        sandbox.wait_on_state(
            self.client.action_runs,
            action_run_url,
            actionrun.ActionRun.FAILED,
        )

        action_run_url = self.client.get_url('MASTER.failjob.1.cleanup')
        sandbox.wait_on_state(
            self.client.action_runs,
            action_run_url,
            actionrun.ActionRun.SUCCEEDED,
        )
        job_runs = self.client.job(self.client.get_url('MASTER.failjob'),
                                   )['runs']
        assert_gt(len(job_runs), 1)

    def test_skip_failed_actions(self):
        config = BASIC_CONFIG + dedent(
            """
        jobs:
          - name: "multi_step_job"
            node: local
            schedule: "daily 04:20"
            actions:
              - name: "broken"
                command: "failingcommand"
              - name: "works"
                command: "echo ok"
                requires: [broken]
        """
        )
        self.start_with_config(config)
        action_run_url = self.client.get_url('MASTER.multi_step_job.0.broken')
        waiter = sandbox.build_waiter_func(
            self.client.action_runs,
            action_run_url,
        )

        waiter(actionrun.ActionRun.FAILED)
        self.sandbox.tronctl('skip', 'MASTER.multi_step_job.0.broken')
        waiter(actionrun.ActionRun.SKIPPED)

        action_run_url = self.client.get_url('MASTER.multi_step_job.0.works')
        sandbox.wait_on_state(
            self.client.action_runs,
            action_run_url,
            actionrun.ActionRun.SUCCEEDED,
        )

        job_run_url = self.client.get_url('MASTER.multi_step_job.0')
        sandbox.wait_on_state(
            self.client.job_runs,
            job_run_url,
            actionrun.ActionRun.SUCCEEDED,
        )

    def test_failure_on_multi_step_job_doesnt_wedge_tron(self):
        config = BASIC_CONFIG + dedent(
            """
            jobs:
                -   name: "random_failure_job"
                    node: local
                    queueing: true
                    schedule: "daily 04:20"
                    actions:
                        -   name: "fa"
                            command: "sleep 0.1; failplz"
                        -   name: "sa"
                            command: "echo 'you will never see this'"
                            requires: [fa]
        """
        )
        self.start_with_config(config)
        job_url = self.client.get_url('MASTER.random_failure_job')

        def wait_on_random_failure_job():
            return len(self.client.job(job_url)['runs']) >= 4

        sandbox.wait_on_sandbox(wait_on_random_failure_job)

        job_runs = self.client.job(job_url)['runs']
        expected = [actionrun.ActionRun.FAILED for _ in range(3)]
        assert_equal([run['state'] for run in job_runs[-3:]], expected)

    def test_cancel_schedules_a_new_run(self):
        config = BASIC_CONFIG + dedent(
            """
            jobs:
                -   name: "a_job"
                    node: local
                    schedule: "daily 05:00:00"
                    actions:
                        -   name: "first_action"
                            command: "echo OK"
        """
        )
        self.start_with_config(config)
        job_name = 'MASTER.a_job'
        job_url = self.client.get_url(job_name)

        self.sandbox.tronctl('cancel', '%s.0' % job_name)

        def wait_on_cancel():
            return len(self.client.job(job_url)['runs']) == 2

        sandbox.wait_on_sandbox(wait_on_cancel)

        run_states = [run['state'] for run in self.client.job(job_url)['runs']]
        expected = [
            actionrun.ActionRun.SCHEDULED,
            actionrun.ActionRun.CANCELLED,
        ]
        assert_equal(run_states, expected)

    def test_job_queueing_false_with_overlap(self):
        """Test that a job that has queueing false properly cancels an
        overlapping job run.
        """
        config = BASIC_CONFIG + dedent(
            """
            jobs:
                -   name: "cancel_overlap"
                    schedule: "cron * * * * *"
                    queueing: False
                    node: local
                    actions:
                        -   name: "do_something"
                            command: "sleep 3s"
                        -   name: "do_other"
                            command: "sleep 3s"
                    cleanup_action:
                        command: "echo done"
        """
        )
        self.start_with_config(config)
        job_url = self.client.get_url('MASTER.cancel_overlap')
        job_run_url = self.client.get_url('MASTER.cancel_overlap.1')

        def wait_on_job_schedule():
            return len(self.client.job(job_url)['runs']) == 2

        sandbox.wait_on_sandbox(wait_on_job_schedule)

        sandbox.wait_on_state(
            self.client.job,
            job_run_url,
            actionrun.ActionRun.CANCELLED,
        )

        action_run_states = [
            action_run['state']
            for action_run in self.client.job_runs(job_run_url)['runs']
        ]
        expected = [
            actionrun.ActionRun.CANCELLED
            for _ in range(len(action_run_states))
        ]
        assert_equal(action_run_states, expected)

    def test_trond_restart_job_with_run_history(self):
        config = BASIC_CONFIG + textwrap.dedent(
            """
           jobs:
              - name: fast_job
                node: local
                schedule: daily 04:20
                actions:
                  - name: single_act
                    command: "sleep 20 && echo good"
        """
        )
        self.start_with_config(config)

        action_run_url = self.client.get_url('MASTER.fast_job.0.single_act')
        sandbox.wait_on_state(
            self.client.action_runs,
            action_run_url,
            actionrun.ActionRun.RUNNING,
        )

        self.restart_trond()

        assert_equal(
            self.client.job_runs(action_run_url)['state'],
            actionrun.ActionRun.UNKNOWN,
        )

        next_run_url = self.client.get_url('MASTER.fast_job.-1.single_act')
        sandbox.wait_on_state(
            self.client.action_runs,
            next_run_url,
            actionrun.ActionRun.RUNNING,
        )

    def test_trond_restart_job_running_with_dependencies(self):
        config = BASIC_CONFIG + textwrap.dedent(
            """
            jobs:
                - name: complex_job
                  node: local
                  schedule: cron * * * * *
                  actions:
                    - name: first_act
                      command: sleep 20 && echo "I'm waiting"
                    - name: following_act
                      command: echo "thing"
                      requires: ['first_act']
                    - name: last_act
                      command: echo foo
                      requires: ['following_act']
        """
        )
        self.start_with_config(config)
        job_name = 'MASTER.complex_job'
        self.sandbox.tronctl('start', job_name)

        action_run_url = self.client.get_url('MASTER.complex_job.1.first_act')
        sandbox.wait_on_state(
            self.client.action_runs,
            action_run_url,
            actionrun.ActionRun.RUNNING,
        )

        self.restart_trond()

        assert_equal(
            self.client.job_runs(action_run_url)['state'],
            actionrun.ActionRun.UNKNOWN,
        )

        for followup_action_run in ('following_act', 'last_act'):
            url = self.client.get_url(
                '%s.1.%s' % (
                    job_name,
                    followup_action_run,
                )
            )
            assert_equal(
                self.client.action_runs(url)['state'],
                actionrun.ActionRun.QUEUED,
            )



import os
import tempfile
from collections import defaultdict

import mock

from testifycompat import assert_equal
from testifycompat import setup
from testifycompat import teardown
from testifycompat import TestCase
from tron.eventbus import EventBus


class MakeEventBusTestCase(TestCase):
    @setup
    def setup(self):
        self.logdir = tempfile.TemporaryDirectory()

    @teardown
    def teardown(self):
        EventBus.shutdown()
        self.logdir.cleanup()

    @mock.patch('tron.eventbus.time', autospec=True)
    def test_setup_eventbus_dir(self, time):
        os.rmdir(self.logdir.name)

        time.time = mock.Mock(return_value=1.0)
        eb = EventBus.create(self.logdir.name)
        assert os.path.exists(self.logdir.name)
        assert os.path.exists(os.path.join(self.logdir.name, "current"))

        time.time = mock.Mock(return_value=2.0)
        eb.event_log = {'foo': 'bar'}
        eb.sync_save_log("test")

        new_eb = EventBus.create(self.logdir.name)
        new_eb.sync_load_log()
        assert new_eb.event_log == eb.event_log


class EventBusTestCase(TestCase):
    @setup
    def setup(self):
        self.log_dir = tempfile.TemporaryDirectory(prefix="tron_eventbus_test")
        self.eventbus = EventBus.create(self.log_dir.name)
        self.eventbus.enabled = True

    @teardown
    def teardown(self):
        EventBus.shutdown()
        self.log_dir.cleanup()

    @mock.patch('tron.eventbus.reactor', autospec=True)
    def test_start(self, reactor):
        self.eventbus.sync_load_log = mock.Mock()
        reactor.callLater = mock.Mock()
        self.eventbus.start()
        assert self.eventbus.sync_load_log.call_count == 1
        assert reactor.callLater.call_count == 1

    def test_shutdown(self):
        assert self.eventbus.enabled
        self.eventbus.sync_save_log = mock.Mock()
        self.eventbus.shutdown()
        assert not self.eventbus.enabled
        assert self.eventbus.sync_save_log.call_count == 1

    def test_publish(self):
        evt = {'id': 'foo'}
        self.eventbus.publish(evt)
        assert self.eventbus.publish_queue.pop() == evt

    def test_subscribe(self):
        ps = ('foo', 'bar', 'cb')
        self.eventbus.subscribe(*ps)
        assert self.eventbus.subscribe_queue.pop() == ps

    def test_has_event(self):
        assert not self.eventbus.has_event('foo')
        self.eventbus.event_log['foo'] = 'bar'
        assert self.eventbus.has_event('foo')

    @mock.patch('tron.eventbus.time', autospec=True)
    def test_sync_load_log(self, time):
        time.time = mock.Mock(return_value=1.0)
        self.eventbus.event_log = {'foo': 'bar'}
        self.eventbus.sync_save_log("test")
        self.eventbus.event_log = {}
        self.eventbus.sync_load_log()
        assert self.eventbus.event_log == {'foo': 'bar'}

    @mock.patch('tron.eventbus.time', autospec=True)
    def test_sync_save_log_time(self, time):
        time.time = mock.Mock(return_value=1.0)
        self.eventbus.sync_save_log("test")
        current_link = os.readlink(self.eventbus.log_current)
        assert_equal(current_link, os.path.join(self.log_dir.name, "1.pickle"))

        time.time = mock.Mock(return_value=2.0)
        self.eventbus.sync_save_log("test")
        new_link = os.readlink(self.eventbus.log_current)
        assert_equal(new_link, os.path.join(self.log_dir.name, "2.pickle"))
        assert os.path.exists(current_link)
        assert os.path.exists(new_link)

    @mock.patch('tron.eventbus.time', autospec=True)
    @mock.patch('tron.eventbus.reactor', autospec=True)
    def test_sync_loop(self, reactor, time):
        time.time = mock.Mock(return_value=0)
        reactor.callLater = mock.Mock()
        self.eventbus.enabled = True
        self.eventbus.sync_shutdown = mock.Mock()
        self.eventbus.sync_loop()
        assert reactor.callLater.call_count == 1
        assert self.eventbus.sync_shutdown.call_count == 0

    @mock.patch('tron.eventbus.reactor', autospec=True)
    def test_sync_loop_shutdown(self, reactor):
        reactor.callLater = mock.Mock()
        self.eventbus.enabled = False
        self.eventbus.sync_save_log = mock.Mock()
        self.eventbus.sync_loop()
        assert reactor.callLater.call_count == 0

    @mock.patch('tron.eventbus.time', autospec=True)
    def test_sync_process_save_log(self, time):
        time.time = mock.Mock(return_value=10)
        self.eventbus.log_updates = 1
        self.eventbus.log_last_save = 0
        self.eventbus.log_save_interval = 20
        self.eventbus.sync_save_log = mock.Mock()
        self.eventbus.sync_process()
        assert self.eventbus.sync_save_log.call_count == 0

        time.time = mock.Mock(return_value=21)
        self.eventbus.sync_process()
        assert self.eventbus.sync_save_log.call_count == 1

        time.time = mock.Mock(return_value=0)
        self.eventbus.log_updates = 0
        self.eventbus.log_save_updates = 20
        self.eventbus.sync_save_log = mock.Mock()
        self.eventbus.sync_process()
        assert self.eventbus.sync_save_log.call_count == 0

        self.eventbus.log_updates = 21
        self.eventbus.sync_process()
        assert self.eventbus.sync_save_log.call_count == 1
        assert self.eventbus.log_updates == 0

    @mock.patch('tron.eventbus.time', autospec=True)
    def test_sync_process_flush_queues(self, time):
        time.time = mock.Mock(return_value=10)
        self.eventbus.sync_subscribe = mock.Mock()
        self.eventbus.sync_publish = mock.Mock()

        for _ in range(5):
            self.eventbus.publish_queue.append(mock.Mock())
            self.eventbus.subscribe_queue.append(mock.Mock())

        self.eventbus.sync_process()

        assert_equal(self.eventbus.sync_subscribe.call_count, 5)
        assert_equal(self.eventbus.sync_publish.call_count, 5)

    @mock.patch('tron.eventbus.reactor', autospec=True)
    def test_sync_publish(self, reactor):
        reactor.callLater = mock.Mock()
        evt = {'id': 'foo', 'bar': 'baz'}
        self.eventbus.event_log = {}
        self.eventbus.log_save_updates = 0
        self.eventbus.sync_publish(evt)
        assert self.eventbus.log_updates == 1
        assert reactor.callLater.call_count == 1

    @mock.patch('tron.eventbus.reactor', autospec=True)
    def test_sync_publish_replace(self, reactor):
        evt1 = {'id': 'foo', 'bar': 'baz'}
        evt2 = {'id': 'foo', 'bar': 'quux'}
        self.eventbus.event_log = {}
        self.eventbus.log_save_updates = 0
        self.eventbus.sync_publish(evt1)
        self.eventbus.sync_publish(evt2)
        assert self.eventbus.log_updates == 2
        assert reactor.callLater.call_count == 2

    @mock.patch('tron.eventbus.reactor', autospec=True)
    def test_sync_publish_duplicate(self, reactor):
        evt = {'id': 'foo', 'bar': 'baz'}
        self.eventbus.event_log = {'foo': {'bar': 'baz'}}
        self.eventbus.log_save_updates = 0
        self.eventbus.sync_publish(evt)
        assert self.eventbus.log_updates == 0
        assert reactor.callLater.call_count == 0

    def test_sync_subscribe(self):
        self.eventbus.event_subscribers = defaultdict(list)
        self.eventbus.sync_subscribe(('pre', 'sub', 'cb'))
        assert self.eventbus.event_subscribers == {'pre': [('sub', 'cb')]}

        self.eventbus.sync_subscribe(('pre', 'sub2', 'cb2'))
        assert self.eventbus.event_subscribers == {
            'pre': [('sub', 'cb'), ('sub2', 'cb2')]
        }

    def test_sync_unsubscribe(self):
        self.eventbus.event_subscribers = defaultdict(list)
        self.eventbus.sync_subscribe(('pre', 'sub', 'cb'))
        self.eventbus.sync_subscribe(('pre', 'sub2', 'cb2'))
        assert self.eventbus.event_subscribers == {
            'pre': [('sub', 'cb'), ('sub2', 'cb2')]
        }

        self.eventbus.sync_unsubscribe(('pre', 'sub'))
        assert self.eventbus.event_subscribers == {'pre': [('sub2', 'cb2')]}
        self.eventbus.sync_unsubscribe(('pre', 'sub2'))
        assert self.eventbus.event_subscribers == {}

    @mock.patch('tron.eventbus.reactor', autospec=True)
    def test_sync_notify(self, reactor):
        reactor.callLater = mock.Mock()
        self.eventbus.event_log = {'p': {}, 'pre': {}, 'prefix': {}}
        self.eventbus.event_subscribers = {
            'pre': [('sub', 'm1')],
            'prefix': [('sub', 'm2'), ('sub2', 'm3')]
        }

        self.eventbus.sync_notify('p')
        assert reactor.callLater.call_count == 0

        self.eventbus.sync_notify('pre')
        assert reactor.callLater.call_count == 1

        self.eventbus.sync_notify('prefix')
        assert reactor.callLater.call_count == 4



import time

import mock
import pytest
from mock import patch
from mock import PropertyMock

from testifycompat import assert_equal
from testifycompat import TestCase
from tron.bin import check_tron_jobs
from tron.bin.check_tron_jobs import State


@pytest.fixture(autouse=True)
def mock_run_interval():
    with patch.object(check_tron_jobs, '_run_interval', 300):
        yield


class TestCheckJobs(TestCase):
    @patch('tron.bin.check_tron_jobs.check_job_result', autospec=True)
    @patch('tron.bin.check_tron_jobs.Client', autospec=True)
    @patch('tron.bin.check_tron_jobs.cmd_utils', autospec=True)
    @patch('tron.bin.check_tron_jobs.parse_cli', autospec=True)
    def test_check_job_result_exception(
        self,
        mock_args,
        mock_cmd_utils,
        mock_client,
        mock_check_job_result,
    ):
        type(mock_args.return_value).job = PropertyMock(return_value=None)
        mock_client.return_value.jobs.return_value = [
            {
                'name': 'job1',
            },
            {
                'name': 'job2',
            },
            {
                'name': 'job3',
            },
        ]
        mock_check_job_result.side_effect = [
            KeyError('foo'),
            None,
            TypeError,
        ]
        error_code = check_tron_jobs.main()
        assert_equal(error_code, 1)
        assert_equal(mock_check_job_result.call_count, 3)

    # These tests test job run succeeded scenarios
    def test_job_succeeded(self):
        job_runs = {
            'status':
                'running',
            'next_run':
                None,
            'runs': [
                {
                    'id':
                        'MASTER.test.3',
                    'state':
                        'scheduled',
                    'run_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() + 600),
                        ),
                    'end_time':
                        None,
                },
                {
                    'id':
                        'MASTER.test.2',
                    'state':
                        'running',
                    'run_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 600),
                        ),
                    'end_time':
                        None,
                },
                {
                    'id':
                        'MASTER.test.1',
                    'state':
                        'succeeded',
                    'run_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 1800),
                        ),
                    'end_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 1700),
                        ),
                },
            ],
        }
        run, state = check_tron_jobs.get_relevant_run_and_state(job_runs)
        assert_equal(run['id'], 'MASTER.test.1')
        assert_equal(state, State.SUCCEEDED)

    def test_job_running_and_action_succeeded(self):
        job_runs = {
            'status':
                'running',
            'next_run':
                None,
            'runs': [
                {
                    'id':
                        'MASTER.test.3',
                    'state':
                        'scheduled',
                    'run_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() + 600),
                        ),
                    'end_time':
                        None,
                },
                {
                    'id':
                        'MASTER.test.2',
                    'state':
                        'running',
                    'run_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 600),
                        ),
                    'end_time':
                        None,
                    'runs': [
                        {
                            'id': 'MASTER.test.2.action2',
                            'state': 'running',
                        },
                        {
                            'id': 'MASTER.test.1.action1',
                            'state': 'succeeded',
                        },
                    ],  # noqa: E122
                },
                {
                    'id':
                        'MASTER.test.1',
                    'state':
                        'succeeded',
                    'run_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 1800),
                        ),
                    'end_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 1700),
                        ),
                },
            ],
        }
        run, state = check_tron_jobs.get_relevant_run_and_state(job_runs)
        assert_equal(run['id'], 'MASTER.test.1')
        assert_equal(state, State.SUCCEEDED)

    def test_get_relevant_action_picks_the_first_one_succeeded(self):
        action_runs = [
            {
                'id':
                    'MASTER.test.action1',
                'action_name':
                    'action1',
                'state':
                    'succeeded',
                'start_time':
                    time.strftime(
                        '%Y-%m-%d %H:%M:%S',
                        time.localtime(time.time() - 1200)
                    ),
                'duration':
                    '0:18:01.475067',
            },
            {
                'id':
                    'MASTER.test.action2',
                'action_name':
                    'action2',
                'state':
                    'succeeded',
                'start_time':
                    time.strftime(
                        '%Y-%m-%d %H:%M:%S', time.localtime(time.time() - 600)
                    ),
                'duration':
                    '0:08:02.005783',
            },
            {
                'id':
                    'MASTER.test.action1',
                'action_name':
                    'action1',
                'state':
                    'succeeded',
                'start_time':
                    time.strftime(
                        '%Y-%m-%d %H:%M:%S', time.localtime(time.time())
                    ),
                'duration':
                    '0:00:01.006305',
            },
        ]
        actual = check_tron_jobs.get_relevant_action(
            action_runs=action_runs,
            last_state=State.SUCCEEDED,
            actions_expected_runtime={
                'action1': 86400.0,
                'action2': 86400.0,
                'action3': 86400.0
            },
        )
        assert_equal(actual["id"], "MASTER.test.action1")

    # These tests test job run failed scenarios
    def test_job_failed(self):
        job_runs = {
            'status':
                'running',
            'next_run':
                None,
            'runs': [
                {
                    'id':
                        'MASTER.test.3',
                    'state':
                        'scheduled',
                    'run_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() + 600),
                        ),
                    'end_time':
                        None,
                },
                {
                    'id':
                        'MASTER.test.2',
                    'state':
                        'running',
                    'run_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 600),
                        ),
                    'end_time':
                        None,
                },
                {
                    'id':
                        'MASTER.test.1',
                    'state':
                        'failed',
                    'run_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 1800),
                        ),
                    'end_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 1700),
                        ),
                },
            ],
        }
        run, state = check_tron_jobs.get_relevant_run_and_state(job_runs)
        assert_equal(run['id'], 'MASTER.test.1')
        assert_equal(state, State.FAILED)

    def test_most_recent_end_time_job_failed(self):
        job_runs = {
            'status':
                'scheduled',
            'next_run':
                None,
            'runs': [
                {
                    'id':
                        'MASTER.test.3',
                    'state':
                        'scheduled',
                    'run_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() + 600),
                        ),
                    'end_time':
                        None
                },
                {
                    'id':
                        'MASTER.test.2',
                    'state':
                        'succeeded',
                    'run_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 1800),
                        ),
                    'end_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 1700),
                        ),
                },
                {
                    'id':
                        'MASTER.test.1',
                    'state':
                        'failed',
                    'run_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 600),
                        ),
                    'end_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 500),
                        ),
                },
            ],
        }
        run, state = check_tron_jobs.get_relevant_run_and_state(job_runs)
        assert_equal(run['id'], 'MASTER.test.1')
        assert_equal(state, State.FAILED)

    def test_rerun_job_failed(self):
        job_runs = {
            'status':
                'scheduled',
            'next_run':
                None,
            'runs': [
                {
                    'id':
                        'MASTER.test.4',
                    'state':
                        'scheduled',
                    'run_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() + 600),
                        ),
                    'end_time':
                        None,
                },
                {
                    'id':
                        'MASTER.test.3',
                    'state':
                        'failed',
                    'run_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 1800),
                        ),
                    'end_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 100),
                        ),
                },
                {
                    'id':
                        'MASTER.test.2',
                    'state':
                        'succeeded',
                    'run_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 600),
                        ),
                    'end_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 500),
                        ),
                },
                {
                    'id':
                        'MASTER.test.1',
                    'state':
                        'failed',
                    'run_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 1800),
                        ),
                    'end_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 1700),
                        ),
                },
            ],
        }
        run, state = check_tron_jobs.get_relevant_run_and_state(job_runs)
        assert_equal(run['id'], 'MASTER.test.3')
        assert_equal(state, State.FAILED)

    def test_job_running_but_action_failed_already(self):
        job_runs = {
            'status':
                'running',
            'next_run':
                None,
            'runs': [
                {
                    'id':
                        'MASTER.test.3',
                    'state':
                        'scheduled',
                    'run_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() + 600),
                        ),
                    'end_time':
                        None,
                },
                {
                    'id':
                        'MASTER.test.2',
                    'state':
                        'running',
                    'run_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 600),
                        ),
                    'end_time':
                        None,
                    'runs': [
                        {
                            'id': 'MASTER.test.2.action2',
                            'state': 'running',
                        },
                        {
                            'id': 'MASTER.test.1.action1',
                            'state': 'failed',
                        },
                    ],  # noqa: E122
                },
                {
                    'id':
                        'MASTER.test.1',
                    'state':
                        'succeeded',
                    'run_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 1800),
                        ),
                    'end_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 1700),
                        ),
                },
            ],
        }
        run, state = check_tron_jobs.get_relevant_run_and_state(job_runs)
        assert_equal(run['id'], 'MASTER.test.2')
        assert_equal(state, State.FAILED)

    def test_get_relevant_action_picks_the_one_that_failed(self):
        action_runs = [
            {
                'node': {
                    'username': 'batch',
                    'hostname': 'localhost',
                    'name': 'localhost',
                    'port': 22,
                },
                'raw_command': '/bin/false',
                'requirements': [],
                'run_num': '582',
                'exit_status': 1,
                'stdout': None,
                'start_time': '2018-02-05 17:40:00',
                'id': 'MASTER.kwatest.582.action1',
                'action_name': 'action1',
                'state': 'failed',
                'command': '/bin/false',
                'end_time': '2018-02-05 17:40:00',
                'stderr': None,
                'duration': '0:00:00.065018',
                'job_name': 'MASTER.kwatest',
            },
            {
                'node': {
                    'username': 'batch',
                    'hostname': 'localhost',
                    'name': 'localhost',
                    'port': 22,
                },
                'raw_command': '/bin/true',
                'requirements': [],
                'run_num': '582',
                'exit_status': 0,
                'stdout': None,
                'start_time': '2018-02-05 17:40:00',
                'id': 'MASTER.kwatest.582.action2',
                'action_name': 'action2',
                'state': 'succeeded',
                'command': '/bin/true',
                'end_time': '2018-02-05 17:40:00',
                'stderr': None,
                'duration': '0:00:00.046243',
                'job_name': 'MASTER.kwatest',
            },
        ]
        actual = check_tron_jobs.get_relevant_action(
            action_runs=action_runs,
            last_state=State.FAILED,
            actions_expected_runtime={}
        )
        assert_equal(actual["state"], "failed")

    # These tests test job/action stuck scenarios
    def test_job_next_run_starting_no_overlap_is_stuck(self):
        job_runs = {
            'status':
                'running',
            'next_run':
                None,
            'runs': [
                {
                    'id':
                        'MASTER.test.2',
                    'state':
                        'queued',
                    'run_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 600),
                        ),
                    'end_time':
                        None,
                },
                {
                    'id':
                        'MASTER.test.1',
                    'state':
                        'running',
                    'run_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 1200),
                        ),
                    'end_time':
                        None,
                },
            ],
        }
        run, state = check_tron_jobs.get_relevant_run_and_state(job_runs)
        assert_equal(run['id'], 'MASTER.test.1')
        assert_equal(state, State.STUCK)

    def test_job_next_run_starting_overlap_allowed_not_stuck(self):
        job_runs = {
            'status':
                'running',
            'next_run':
                None,
            'allow_overlap':
                True,
            'runs': [
                {
                    'id':
                        'MASTER.test.3',
                    'state':
                        'queued',
                    'run_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 600),
                        ),
                    'end_time':
                        None,
                },
                {
                    'id':
                        'MASTER.test.2',
                    'state':
                        'running',
                    'run_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 1200),
                        ),
                    'end_time':
                        None,
                },
                {
                    'id':
                        'MASTER.test.1',
                    'state':
                        'succeeded',
                    'run_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 1800),
                        ),
                    'end_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 1700),
                        ),
                },
            ],
        }
        run, state = check_tron_jobs.get_relevant_run_and_state(job_runs)
        assert_equal(run['id'], 'MASTER.test.1')
        assert_equal(state, State.SUCCEEDED)

    def test_job_running_job_exceeds_expected_runtime(self):
        job_runs = {
            'status':
                'running',
            'next_run':
                None,
            'expected_runtime':
                480.0,
            'allow_overlap':
                True,
            'runs': [
                {
                    'id':
                        'MASTER.test.100',
                    'state':
                        'scheduled',
                    'run_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() + 600),
                        ),
                    'end_time':
                        None,
                    'start_time':
                        None,
                    'duration':
                        '',
                },
                {
                    'id':
                        'MASTER.test.99',
                    'state':
                        'running',
                    'run_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 600),
                        ),
                    'start_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 600),
                        ),
                    'end_time':
                        None,
                    'duration':
                        '0:10:01.883601',
                },
            ],
        }
        run, state = check_tron_jobs.get_relevant_run_and_state(job_runs)
        assert_equal(run['id'], 'MASTER.test.99')
        assert_equal(state, State.STUCK)

    def test_job_running_action_exceeds_expected_runtime(self):
        job_runs = {
            'status':
                'running',
            'next_run':
                None,
            'actions_expected_runtime': {
                'action1': 720.0,
                'action2': 480.0
            },
            'runs': [
                dict(
                    id='MASTER.test.3',
                    state='scheduled',
                    run_time=time.strftime(
                        '%Y-%m-%d %H:%M:%S',
                        time.localtime(time.time() + 600),
                    ),
                    end_time=None,
                    duration='',
                ),
                dict(
                    id='MASTER.test.2',
                    state='running',
                    run_time=time.strftime(
                        '%Y-%m-%d %H:%M:%S',
                        time.localtime(time.time() - 600),
                    ),
                    end_time=None,
                    duration='0:10:01.883601',
                    runs=[
                        dict(
                            id='MASTER.test.2.action2',
                            state='running',
                            action_name='action2',
                            start_time=time.strftime(
                                '%Y-%m-%d %H:%M:%S',
                                time.localtime(time.time() - 600),
                            ),
                            duration='0:10:01.883601',
                        ),
                        dict(
                            id='MASTER.test.2.action1',
                            state='running',
                            action_name='action1',
                            start_time=time.strftime(
                                '%Y-%m-%d %H:%M:%S',
                                time.localtime(time.time() - 600),
                            ),
                            duration='0:10:01.885401',
                        ),
                    ],
                ),
                dict(
                    id='MASTER.test.1',
                    state='succeeded',
                    run_time=time.strftime(
                        '%Y-%m-%d %H:%M:%S',
                        time.localtime(time.time() - 1800),
                    ),
                    end_time=time.strftime(
                        '%Y-%m-%d %H:%M:%S',
                        time.localtime(time.time() - 1700),
                    ),
                    duration='0:15:00.453601',
                ),
            ],
        }
        run, state = check_tron_jobs.get_relevant_run_and_state(job_runs)
        assert_equal(run['id'], 'MASTER.test.2')
        assert_equal(state, State.STUCK)

    def test_job_running_action_exceeds_expected_runtime_and_other_action_failed(self):
        job_runs = {
            'status':
                'running',
            'next_run':
                None,
            'actions_expected_runtime': {
                'action1': 720.0,
                'action2': 480.0
            },
            'runs': [
                dict(
                    id='MASTER.test.1',
                    state='running',
                    run_time=time.strftime(
                        '%Y-%m-%d %H:%M:%S',
                        time.localtime(time.time() - 600),
                    ),
                    end_time=None,
                    duration='0:10:01.883601',
                    runs=[
                        dict(
                            id='MASTER.test.1.action2',
                            state='failed',
                            action_name='action2',
                            start_time=time.strftime(
                                '%Y-%m-%d %H:%M:%S',
                                time.localtime(time.time() - 600),
                            ),
                            duration='0:10:01.883601',
                        ),
                        dict(
                            id='MASTER.test.1.action1',
                            state='running',
                            action_name='action1',
                            start_time=time.strftime(
                                '%Y-%m-%d %H:%M:%S',
                                time.localtime(time.time() - 600),
                            ),
                            duration='0:10:01.885401',
                        ),
                    ],
                ),
            ],
        }
        run, state = check_tron_jobs.get_relevant_run_and_state(job_runs)
        assert_equal(run['id'], 'MASTER.test.1')
        assert_equal(state, State.FAILED)

    def test_job_stuck_when_runtime_not_sorted(self):
        job_runs = {
            'status':
                'running',
            'next_run':
                None,
            'runs': [
                {
                    'id':
                        'MASTER.test.2',
                    'state':
                        'running',
                    'run_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 600),
                        ),
                    'end_time':
                        None,
                },
                {
                    'id':
                        'MASTER.test.1',
                    'state':
                        'scheduled',
                    'run_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time()),
                        ),
                    'end_time':
                        None,
                },
            ],
        }
        run, state = check_tron_jobs.get_relevant_run_and_state(job_runs)
        assert_equal(run['id'], 'MASTER.test.2')
        assert_equal(state, State.STUCK)

    def test_get_relevant_action_pick_the_one_stuck(self):
        action_runs = [
            {
                'id':
                    'MASTER.test.1.action3',
                'state':
                    'succeeded',
                'start_time':
                    time.strftime(
                        '%Y-%m-%d %H:%M:%S',
                        time.localtime(time.time() - 1200)
                    ),
                'duration':
                    '0:18:01.475067',
            },
            {
                'id':
                    'MASTER.test.1.action2',
                'state':
                    'running',
                'start_time':
                    time.strftime(
                        '%Y-%m-%d %H:%M:%S',
                        time.localtime(time.time() - 1100)
                    ),
                'duration':
                    '0:18:40.005783',
            },
            {
                'id':
                    'MASTER.test.1.action1',
                'state':
                    'succeeded',
                'start_time':
                    time.strftime(
                        '%Y-%m-%d %H:%M:%S',
                        time.localtime(time.time() - 1000)
                    ),
                'duration':
                    '0:00:01.006305',
            },
        ]
        actual = check_tron_jobs.get_relevant_action(
            action_runs=action_runs,
            last_state=State.STUCK,
            actions_expected_runtime={
                'action1': 86400.0,
                'action2': 86400.0,
                'action3': 86400.0,
            }
        )
        assert_equal(actual["id"], "MASTER.test.1.action2")

    def test_get_relevant_action_pick_the_one_exceeds_expected_runtime(self):
        action_runs = [
            {
                'id':
                    'MASTER.test.1.action3',
                'state':
                    'running',
                'start_time':
                    time.strftime(
                        '%Y-%m-%d %H:%M:%S', time.localtime(time.time() - 600)
                    ),
                'duration':
                    '0:10:00.006305',
            },
            {
                'id':
                    'MASTER.test.1.action2',
                'state':
                    'running',
                'start_time':
                    time.strftime(
                        '%Y-%m-%d %H:%M:%S', time.localtime(time.time() - 600)
                    ),
                'duration':
                    '0:10:00.006383',
            },
            {
                'id':
                    'MASTER.test.1.action1',
                'state':
                    'succeeded',
                'start_time':
                    time.strftime(
                        '%Y-%m-%d %H:%M:%S', time.localtime(time.time() - 600)
                    ),
                'duration':
                    '0:10:00.006331',
            },
        ]
        actions_expected_runtime = {
            'action3': 480.0,
            'action2': 720.0,
            'action1': 900.0
        }
        actual = check_tron_jobs.get_relevant_action(
            action_runs=action_runs,
            last_state=State.STUCK,
            actions_expected_runtime=actions_expected_runtime
        )
        assert_equal(actual["id"], "MASTER.test.1.action3")

    def test_get_relevant_action_pick_the_one_exceeds_expected_runtime_with_long_duration(
        self
    ):
        action_runs = [
            {
                'id':
                    'MASTER.test.1.action3',
                'action_name':
                    'action3',
                'state':
                    'running',
                'start_time':
                    time.strftime(
                        '%Y-%m-%d %H:%M:%S', time.localtime(time.time() - 600)
                    ),
                'duration':
                    '1 day, 0:10:00.006305',
            },
            {
                'id':
                    'MASTER.test.1.action2',
                'action_name':
                    'action2',
                'state':
                    'running',
                'start_time':
                    time.strftime(
                        '%Y-%m-%d %H:%M:%S', time.localtime(time.time() - 600)
                    ),
                'duration':
                    '2 days, 0:10:00.006383',
            },
            {
                'id':
                    'MASTER.test.1.action1',
                'action_name':
                    'action1',
                'state':
                    'running',
                'start_time':
                    time.strftime(
                        '%Y-%m-%d %H:%M:%S', time.localtime(time.time() - 600)
                    ),
                'duration':
                    '1 day, 0:10:00.006331',
            },
        ]
        actions_expected_runtime = {
            'action3': 100000.0,
            'action2': 100000.0,
            'action1': 100000.0
        }
        actual = check_tron_jobs.get_relevant_action(
            action_runs=action_runs,
            last_state=State.STUCK,
            actions_expected_runtime=actions_expected_runtime
        )
        assert_equal(actual["id"], "MASTER.test.1.action2")

    def test_no_job_scheduled_or_queuing(self):
        """ If the past 2 runs succeeded but no future job is scheuled,
        we should consider the job to have suceeded.
        """
        job_runs = {
            'status':
                'succeeded',
            'next_run':
                None,
            'runs': [
                {
                    'id':
                        'MASTER.test.2',
                    'state':
                        'succeeded',
                    'run_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 300),
                        ),
                    'end_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 600),
                        ),
                },
                {
                    'id':
                        'MASTER.test.1',
                    'state':
                        'succeeded',
                    'run_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 900),
                        ),
                    'end_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 1200),
                        ),
                },
            ],
            'monitoring': {},
        }
        run, state = check_tron_jobs.get_relevant_run_and_state(job_runs)
        assert_equal(run['id'], 'MASTER.test.2')
        assert_equal(state, State.SUCCEEDED)

    # These tests test job without succeeded/failed run scenarios
    def test_job_no_runs_to_check(self):
        job_runs = {
            'status':
                'scheduled',
            'next_run':
                None,
            'runs': [{
                'id':
                    'MASTER.test.1',
                'state':
                    'scheduled',
                'run_time':
                    time.strftime(
                        '%Y-%m-%d %H:%M:%S',
                        time.localtime(time.time() + 1200),
                    ),
                'end_time':
                    None,
            }, ],
        }
        run, state = check_tron_jobs.get_relevant_run_and_state(job_runs)
        assert_equal(run['id'], 'MASTER.test.1')
        assert_equal(state, State.NO_RUNS_TO_CHECK)

    def test_job_has_no_runs_at_all(self):
        job_runs = {
            'status': 'running',
            'next_run': None,
            'runs': [],
        }
        run, state = check_tron_jobs.get_relevant_run_and_state(job_runs)
        assert_equal(run, None)
        assert_equal(state, State.NO_RUN_YET)

    # These tests test job/action unknown scenarios
    def test_job_unknown(self):
        job_runs = {
            'status':
                'running',
            'next_run':
                None,
            'runs': [
                {
                    'id':
                        'MASTER.test.3',
                    'state':
                        'scheduled',
                    'run_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 600),
                        ),
                    'end_time':
                        None,
                },
                {
                    'id':
                        'MASTER.test.2',
                    'state':
                        'unknown',
                    'run_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 1200),
                        ),
                    'end_time':
                        None,
                },
                {
                    'id':
                        'MASTER.test.1',
                    'state':
                        'succeeded',
                    'run_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 1800),
                        ),
                    'end_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 1700),
                        ),
                },
            ],
        }
        run, state = check_tron_jobs.get_relevant_run_and_state(job_runs)
        assert_equal(run['id'], 'MASTER.test.2')
        assert_equal(state, State.UNKNOWN)

    def test_job_running_but_action_unknown_already(self):
        job_runs = {
            'status':
                'running',
            'next_run':
                None,
            'runs': [
                {
                    'id':
                        'MASTER.test.3',
                    'state':
                        'scheduled',
                    'run_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() + 600),
                        ),
                    'end_time':
                        None,
                },
                {
                    'id':
                        'MASTER.test.2',
                    'state':
                        'running',
                    'run_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 600),
                        ),
                    'end_time':
                        None,
                    'runs': [
                        {
                            'id': 'MASTER.test.2.action2',
                            'state': 'running',
                        },
                        {
                            'id': 'MASTER.test.1.action1',
                            'state': 'unknown',
                        },
                    ],  # noqa: E122
                },
                {
                    'id':
                        'MASTER.test.1',
                    'state':
                        'succeeded',
                    'run_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 1800),
                        ),
                    'end_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 1700),
                        ),
                },
            ],
        }
        run, state = check_tron_jobs.get_relevant_run_and_state(job_runs)
        assert_equal(run['id'], 'MASTER.test.2')
        assert_equal(state, State.UNKNOWN)

    def test_job_waiting_but_action_unknown_already(self):
        job_runs = {
            'status':
                'waiting',
            'next_run':
                None,
            'runs': [
                {
                    'id':
                        'MASTER.test.3',
                    'state':
                        'scheduled',
                    'run_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() + 600),
                        ),
                    'end_time':
                        None,
                },
                {
                    'id':
                        'MASTER.test.2',
                    'state':
                        'waiting',
                    'run_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 600),
                        ),
                    'end_time':
                        None,
                    'runs': [
                        {
                            'id': 'MASTER.test.2.action2',
                            'state': 'waiting',
                        },
                        {
                            'id': 'MASTER.test.1.action1',
                            'state': 'unknown',
                        },
                    ],  # noqa: E122
                },
                {
                    'id':
                        'MASTER.test.1',
                    'state':
                        'succeeded',
                    'run_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 1800),
                        ),
                    'end_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 1700),
                        ),
                },
            ],
        }
        run, state = check_tron_jobs.get_relevant_run_and_state(job_runs)
        assert_equal(run['id'], 'MASTER.test.2')
        assert_equal(state, State.UNKNOWN)

    # These tests test guess realert feature
    def test_guess_realert_every(self):
        job_runs = {
            'status':
                'running',
            'next_run':
                time.strftime(
                    '%Y-%m-%d %H:%M:%S',
                    time.localtime(time.time() + 600),
                ),
            'runs': [
                {
                    'id': 'MASTER.test.3',
                    'state': 'scheduled',
                    'start_time': None,
                    'run_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() + 600),
                        ),
                },
                {
                    'id':
                        'MASTER.test.2',
                    'state':
                        'failed',
                    'start_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 600),
                        ),
                },
                {
                    'id':
                        'MASTER.test.1',
                    'state':
                        'succeeded',
                    'start_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 1800),
                        ),
                },
            ],  # noqa: E122
        }
        realert_every = check_tron_jobs.guess_realert_every(job_runs)
        assert_equal(realert_every, 4)

    def test_guess_realert_every_no_action_run_starts(self):
        job_runs = {
            'status':
                'running',
            'next_run':
                time.strftime(
                    '%Y-%m-%d %H:%M:%S',
                    time.localtime(time.time() + 600),
                ),
            'runs': [
                {
                    'id': 'MASTER.test.3',
                    'state': 'scheduled',
                    'start_time': None,
                },
                {
                    'id':
                        'MASTER.test.2',
                    'state':
                        'failed',
                    'start_time': None,
                    'run_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 200),
                        ),
                },
                {
                    'id':
                        'MASTER.test.1',
                    'state':
                        'succeeded',
                    'start_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 600),
                        ),
                    'run_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 600),
                        ),
                },
            ],  # noqa: E122
        }
        realert_every = check_tron_jobs.guess_realert_every(job_runs)
        assert_equal(realert_every, 2)

    def test_guess_realert_every_queue_job(self):
        job_runs = {
            'status':
                'running',
            'next_run':
                None,
            'runs': [
                {
                    'id': 'MASTER.test.3',
                    'state': 'queued',
                    'start_time': None,
                },
                {
                    'id':
                        'MASTER.test.2',
                    'state':
                        'running',
                    'start_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 600),
                        ),
                },
                {
                    'id':
                        'MASTER.test.1',
                    'state':
                        'succeeded',
                    'start_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 1800),
                        ),
                },
            ],
        }
        realert_every = check_tron_jobs.guess_realert_every(job_runs)
        assert_equal(realert_every, -1)

    def test_guess_realert_every_frequent_run(self):
        job_runs = {
            'status':
                'running',
            'next_run':
                time.strftime(
                    '%Y-%m-%d %H:%M:%S',
                    time.localtime(time.time() + 10),
                ),
            'runs': [
                {
                    'id': 'MASTER.test.3',
                    'state': 'scheduled',
                    'start_time': None,
                },
                {
                    'id':
                        'MASTER.test.2',
                    'state':
                        'failed',
                    'start_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 10),
                        ),
                },
                {
                    'id':
                        'MASTER.test.1',
                    'state':
                        'succeeded',
                    'start_time':
                        time.strftime(
                            '%Y-%m-%d %H:%M:%S',
                            time.localtime(time.time() - 20),
                        ),
                },
            ],  # noqa: E122
        }
        realert_every = check_tron_jobs.guess_realert_every(job_runs)
        assert_equal(realert_every, 1)

    def test_guess_realert_every_first_time_job(self):
        job_runs = {
            'status':
                'enabled',
            'next_run':
                time.strftime(
                    '%Y-%m-%d %H:%M:%S',
                    time.localtime(time.time() + 600),
                ),
            'runs': [{
                'id': 'MASTER.test.1',
                'state': 'scheduled',
                'start_time': None,
            }, ],
        }
        realert_every = check_tron_jobs.guess_realert_every(job_runs)
        assert_equal(realert_every, -1)


class TestCheckPreciousJobs:
    @pytest.fixture(autouse=True)
    def setup_job(self):
        self.job_name = 'fake_job'
        self.monitoring = {
            'team': 'fake_team',
            'notification_email': 'fake_email',
            check_tron_jobs.PRECIOUS_JOB_ATTR: True,
        }
        self.runs = [{
            'id': f"{self.job_name}.15",
            'job_name': self.job_name,
            'run_num': 15,
            'run_time': '2018-10-13 12:00:00',
            'start_time': '2018-10-13 12:00:00',
            'end_time': '2018-10-13 12:30:00',
            'state': 'succeeded',
            'exit_status': 0,
        }, {
            'id': f"{self.job_name}.14",
            'job_name': self.job_name,
            'run_num': 14,
            'run_time': '2018-10-12 12:00:00',
            'start_time': '2018-10-12 12:00:00',
            'end_time': '2018-10-12 12:30:00',
            'state': 'failed',
            'exit_status': 1,
        }, {
            'id': f"{self.job_name}.13",
            'job_name': self.job_name,
            'run_num': 13,
            'run_time': '2018-10-11 13:00:00',
            'start_time': '2018-10-11 13:00:00',
            'end_time': '2018-10-11 13:30:00',
            'state': 'failed',
            'exit_status': 1,
        }, {
            'id': f"{self.job_name}.12",
            'job_name': self.job_name,
            'run_num': 12,
            'run_time': '2018-10-11 12:00:00',
            'start_time': '2018-10-11 12:00:00',
            'end_time': '2018-10-11 12:30:00',
            'state': 'succeeded',
            'exit_status': 0,
        }, {
            'id': f"{self.job_name}.11",
            'job_name': self.job_name,
            'run_num': 11,
            'run_time': '2018-10-10 13:00:00',
            'start_time': '2018-10-10 13:00:00',
            'end_time': '2018-10-10 13:30:00',
            'state': 'succeeded',
            'exit_status': 0,
        }, {
            'id': f"{self.job_name}.10",
            'job_name': self.job_name,
            'run_num': 10,
            'run_time': '2018-10-10 12:00:00',
            'start_time': '2018-10-10 12:00:00',
            'end_time': '2018-10-10 12:30:00',
            'state': 'failed',
            'exit_status': 1,
        }]
        self.job = {
            'name': 'fake_job',
            'status': 'enabled',
            'monitoring': self.monitoring,
            'runs': self.runs,
        }

    @patch('time.time', mock.Mock(return_value=1539460800.0), autospec=None)
    def test_sort_runs_by_interval_day(self):
        run_buckets = check_tron_jobs.sort_runs_by_interval(self.job, 'day')

        assert set(run_buckets.keys()) == \
            set(['2018.10.10', '2018.10.11', '2018.10.12', '2018.10.13'])
        assert len(run_buckets['2018.10.10']) == 2
        assert len(run_buckets['2018.10.11']) == 2
        assert len(run_buckets['2018.10.12']) == 1
        assert len(run_buckets['2018.10.13']) == 1

    @patch('time.time', mock.Mock(return_value=1539633600.0), autospec=None)
    def test_sort_runs_by_interval_day_empty_buckets(self):
        self.job['runs'] = [{
            'id': f"{self.job_name}.16",
            'job_name': self.job_name,
            'run_num': 16,
            'run_time': '2018-10-15 12:00:00',
            'start_time': '2018-10-15 12:00:00',
            'end_time': '2018-10-15 12:30:00',
            'state': 'succeeded',
            'exit_status': 0,
        }] + self.job['runs']

        run_buckets = check_tron_jobs.sort_runs_by_interval(self.job, 'day')

        assert '2018.10.14' in run_buckets
        assert run_buckets['2018.10.14'] == []

    @patch('time.time', mock.Mock(return_value=1539633600.0), autospec=None)
    def test_sort_runs_by_interval_day_old_empty_buckets(self):
        # If the newest run is a backfill for an older date, shouldn't be
        # included for buckets
        self.job['runs'] = [{
            'id': f"{self.job_name}.16",
            'job_name': self.job_name,
            'run_num': 16,
            'run_time': '2018-10-01 12:00:00',
            'start_time': '2018-10-01 12:00:00',
            'end_time': '2018-10-01 12:30:00',
            'state': 'succeeded',
            'exit_status': 0,
        }] + self.job['runs']

        run_buckets = check_tron_jobs.sort_runs_by_interval(self.job, 'day')
        # Current time is patched to 2018-10-15, so we should include
        # NUM_PRECIOUS intervals, starting 2018-10-09
        # 2018-10-01 should not be included, even though there is a run for it,
        # because it is too old
        assert '2018.10.01' not in run_buckets
        assert run_buckets['2018.10.14'] == []
        assert len(run_buckets) == 7

    @patch('tron.bin.check_tron_jobs.guess_realert_every', mock.Mock(return_value=1), autospec=None)
    @patch('tron.bin.check_tron_jobs.Client', autospec=True)
    @patch('tron.bin.check_tron_jobs.compute_check_result_for_job_runs', autospec=True)
    @patch('tron.bin.check_tron_jobs.get_object_type_from_identifier', autospec=True)
    def test_compute_check_result_for_job_not_precious(
        self,
        mock_get_obj_type,
        mock_check_job_runs,
        mock_client,
    ):
        client = mock_client('fake_server')
        client.cluster_name = 'fake_cluster'
        del self.job['monitoring'][check_tron_jobs.PRECIOUS_JOB_ATTR]
        client.job = mock.Mock(return_value=self.job)
        mock_check_job_runs.return_value = {
            'output': 'fake_output',
            'status': 'fake_status'
        }

        results = check_tron_jobs.compute_check_result_for_job(
            client, self.job, url_index={},
        )

        # make sure all job runs for a job are included by not incl count arg
        assert client.job.call_args_list == [mock.call(
            mock_get_obj_type.return_value.url,
            include_action_runs=True,
        )]
        assert len(results) == 1
        assert results[0]['name'] == 'check_tron_job.fake_job'
        assert mock_check_job_runs.call_count == 1

    @patch('tron.bin.check_tron_jobs.guess_realert_every', mock.Mock(return_value=1), autospec=None)
    @patch('tron.bin.check_tron_jobs.Client', autospec=True)
    def test_compute_check_result_for_job_disabled(self, mock_client):
        client = mock_client('fake_server')
        client.cluster_name = 'fake_cluster'
        self.job['status'] = 'disabled'

        results = check_tron_jobs.compute_check_result_for_job(
            client, self.job, url_index={},
        )

        assert len(results) == 1
        assert results[0]['status'] == 0
        assert results[0]['output'] == \
            "OK: fake_job is disabled and won't be checked."

    @patch('tron.bin.check_tron_jobs.guess_realert_every', mock.Mock(return_value=1), autospec=None)
    @patch('time.time', mock.Mock(return_value=1539460800.0), autospec=None)
    @patch('tron.bin.check_tron_jobs.Client', autospec=True)
    @patch('tron.bin.check_tron_jobs.compute_check_result_for_job_runs', autospec=True)
    @patch('tron.bin.check_tron_jobs.get_object_type_from_identifier', autospec=True)
    def test_compute_check_result_for_job_enabled(
        self,
        mock_get_obj_type,
        mock_check_job_runs,
        mock_client,
    ):
        client = mock_client('fake_server')
        client.cluster_name = 'fake_cluster'
        self.job['monitoring']['check_every'] = 500
        self.job['monitoring']['hide_stderr'] = True
        client.job = mock.Mock(return_value=self.job)
        mock_check_job_runs.return_value = {
            'output': 'fake_output',
            'status': 'fake_status'
        }

        results = check_tron_jobs.compute_check_result_for_job(
            client, self.job, url_index={},
        )

        # Test that hide_stderr is passed to check_job_runs
        assert mock_check_job_runs.call_args_list[0][1]['hide_stderr'] is True

        # make sure all job runs for a job are included by not incl count arg
        assert client.job.call_args_list == [mock.call(
            mock_get_obj_type.return_value.url,
            include_action_runs=True,
        )]
        assert len(results) == 4
        assert set([res['name'] for res in results]) == set([
            'check_tron_job.fake_job-2018.10.10',
            'check_tron_job.fake_job-2018.10.11',
            'check_tron_job.fake_job-2018.10.12',
            'check_tron_job.fake_job-2018.10.13',
        ])
        for res in results:
            assert res['check_every'] == '300s'
            assert check_tron_jobs.PRECIOUS_JOB_ATTR not in res



from __future__ import absolute_import
from __future__ import unicode_literals

import logging

from tron.core.actionrun import ActionRun
from tron.core.actionrun import MesosActionRun
from tron.core.actionrun import SSHActionRun

log = logging.getLogger(__name__)


def filter_action_runs_needing_recovery(action_runs):
    ssh_runs = []
    mesos_runs = []
    for action_run in action_runs:
        if isinstance(action_run, SSHActionRun):
            if action_run.state == ActionRun.UNKNOWN:
                ssh_runs.append(action_run)
        elif isinstance(action_run, MesosActionRun):
            if action_run.state == ActionRun.UNKNOWN and action_run.end_time is None:
                mesos_runs.append(action_run)
    return ssh_runs, mesos_runs


def launch_recovery_actionruns_for_job_runs(job_runs, master_action_runner):
    for run in job_runs:
        if not run._action_runs:
            log.info(f'Skipping recovery of {run} with no action runs (may have been cleaned up)')
            continue

        ssh_runs, mesos_runs = filter_action_runs_needing_recovery(run._action_runs)
        for action_run in ssh_runs:
            action_run.recover()

        for action_run in mesos_runs:
            action_run.recover()



from __future__ import absolute_import
from __future__ import unicode_literals

from twisted.internet import defer
from twisted.internet import reactor
from twisted.python import failure


class Error(Exception):
    pass


def _cancel(deferred):
    """Re-implementing what's available in newer twisted in a crappy, but
    workable way."""

    if not deferred.called:
        deferred.errback(failure.Failure(Error()))
    elif isinstance(deferred.result, defer.Deferred):
        _cancel(deferred.result)


def defer_timeout(deferred, timeout):
    try:
        reactor.callLater(timeout, deferred.cancel)
    except AttributeError:
        reactor.callLater(timeout, lambda: _cancel(deferred))



import collections
import datetime
from unittest.mock import MagicMock

import mock

from testifycompat import assert_equal
from testifycompat import assert_not_equal
from testifycompat import setup
from testifycompat import setup_teardown
from testifycompat import TestCase
from tests.assertions import assert_call
from tests.assertions import assert_length
from tests.testingutils import autospec_method
from tron import actioncommand
from tron import node
from tron.core import job
from tron.core import jobrun
from tron.core.actionrun import ActionRun
from tron.core.job_scheduler import JobScheduler


class TestJob(TestCase):
    @setup_teardown
    def setup_job(self):
        action_graph = mock.Mock(names=lambda: ['one', 'two'])
        scheduler = mock.Mock()
        run_collection = MagicMock()
        self.nodes = mock.create_autospec(node.NodePool)
        self.action_runner = mock.create_autospec(
            actioncommand.SubprocessActionRunnerFactory,
        )

        patcher = mock.patch(
            'tron.core.job.node.NodePoolRepository', autospec=True
        )
        with patcher as self.mock_node_repo:
            self.job = job.Job(
                "jobname",
                scheduler,
                run_collection=run_collection,
                action_graph=action_graph,
                node_pool=self.nodes,
                action_runner=actioncommand.NoActionRunnerFactory
            )
            autospec_method(self.job.notify)
            autospec_method(self.job.watch)
            yield

    def test__init__(self):
        assert str(self.job.output_path).endswith(self.job.name)

    def test_from_config(self):
        action = mock.MagicMock(
            name='first',
            command='doit',
            node=None,
            requires=[],
        )
        job_config = mock.Mock(
            node='thenodepool',
            monitoring={
                "team": "foo",
                "page": True,
            },
            all_nodes=False,
            queueing=True,
            enabled=True,
            run_limit=20,
            actions={action.name: action},
            cleanup_action=None,
        )
        job_config.name = 'ajob'  # set this after mock creation to give it a "real" name attribute
        scheduler = 'scheduler_token'
        parent_context = 'parent_context_token'
        output_path = ["base_path"]
        new_job = job.Job.from_config(
            job_config,
            scheduler,
            parent_context=parent_context,
            output_path=output_path,
            action_runner=self.action_runner,
            action_graph=mock.Mock()
        )

        assert_equal(new_job.scheduler, scheduler)
        assert_equal(new_job.context.next, parent_context)
        self.mock_node_repo.get_instance().get_by_name.assert_called_with(
            job_config.node,
        )
        assert_equal(new_job.enabled, True)
        assert_equal(new_job.get_monitoring()["team"], "foo")
        assert new_job.action_graph

    def test_update_from_job(self):
        action_runner = mock.Mock()
        other_job = job.Job(
            'otherjob',
            'scheduler',
            action_runner=action_runner,
            run_limit=10,
        )
        self.job.update_from_job(other_job)
        assert_equal(self.job.name, 'otherjob')
        assert_equal(self.job.scheduler, 'scheduler')
        assert_equal(self.job, other_job)
        assert_equal(self.job.runs.run_limit, 10)

    def test_status_disabled(self):
        self.job.enabled = False
        assert_equal(self.job.status, self.job.STATUS_DISABLED)

    def test_status_enabled(self):
        self.job.runs.get_run_by_state = lambda state: MagicMock() if state == ActionRun.SCHEDULED else None
        self.job.runs.get_active.return_value = []
        assert_equal(self.job.status, self.job.STATUS_ENABLED)

    def test_status_running(self):
        self.job.runs.get_active.return_value = [MagicMock()]
        assert_equal(self.job.status, self.job.STATUS_RUNNING)

    def test_status_unknown(self):
        self.job.runs.get_active.return_value = []
        self.job.runs.get_run_by_state = lambda s: None
        assert_equal(self.job.status, self.job.STATUS_UNKNOWN)

    def test_state_data(self):
        state_data = self.job.state_data
        assert_equal(state_data['runs'], self.job.runs.state_data)
        assert state_data['enabled']

    def test_get_job_runs_from_state(self):
        job_runs = [
            dict(
                run_num=i,
                job_name="thename",
                run_time="sometime",
                start_time="start_time",
                end_time="sometime",
                cleanup_run=None,
                runs=[],
            ) for i in range(0, 3)
        ]
        state_data = {'enabled': False, 'runs': job_runs}
        self.job.get_job_runs_from_state(state_data)
        assert not self.job.enabled

    def test_build_new_runs(self):
        run_time = datetime.datetime(2012, 3, 14, 15, 9, 26)
        runs = list(self.job.build_new_runs(run_time))

        self.job.node_pool.next.assert_called_with()
        node = self.job.node_pool.next.return_value
        assert_call(
            self.job.runs.build_new_run,
            0,
            self.job,
            run_time,
            node,
            manual=False,
        )
        assert_length(runs, 1)
        self.job.watch.assert_called_with(runs[0])

    def test_build_new_runs_all_nodes(self):
        self.job.all_nodes = True
        run_time = datetime.datetime(2012, 3, 14, 15, 9, 26)
        node_count = 2
        self.job.node_pool.nodes = [mock.Mock()] * node_count
        runs = list(self.job.build_new_runs(run_time))

        assert_length(runs, node_count)
        for i in range(len(runs)):
            node = self.job.node_pool.nodes[i]
            assert_call(
                self.job.runs.build_new_run,
                i,
                self.job,
                run_time,
                node,
                manual=False,
            )

        calls = []
        for r in runs:
            calls.extend(r.mock_calls)
        self.job.watch.assert_has_calls(calls)

    def test_build_new_runs_manual(self):
        run_time = datetime.datetime(2012, 3, 14, 15, 9, 26)
        runs = list(self.job.build_new_runs(run_time, manual=True))

        self.job.node_pool.next.assert_called_with()
        node = self.job.node_pool.next.return_value
        assert_length(runs, 1)
        assert_call(
            self.job.runs.build_new_run,
            0,
            self.job,
            run_time,
            node,
            manual=True,
        )
        self.job.watch.assert_called_with(runs[0])

    def test_handler(self):
        self.job.handler(None, jobrun.JobRun.NOTIFY_STATE_CHANGED)
        self.job.notify.assert_called_with(self.job.NOTIFY_STATE_CHANGE)

        self.job.handler(None, jobrun.JobRun.NOTIFY_DONE)
        self.job.notify.assert_called_with(self.job.NOTIFY_RUN_DONE)

    def test__eq__(self):
        other_job = job.Job("jobname", 'scheduler', run_collection=MagicMock())
        assert not self.job == other_job
        other_job.update_from_job(self.job)
        assert_equal(self.job, other_job)

    def test__ne__(self):
        other_job = job.Job("jobname", 'scheduler', run_collection=MagicMock())
        assert self.job != other_job
        other_job.update_from_job(self.job)
        assert not self.job != other_job

    def test__eq__true(self):
        action_runner = mock.Mock()
        first = job.Job("jobname", 'scheduler', action_runner=action_runner)
        second = job.Job("jobname", 'scheduler', action_runner=action_runner)
        assert_equal(first, second)

    def test__eq__false(self):
        first = job.Job("jobname", 'scheduler', action_runner=mock.Mock())
        second = job.Job("jobname", 'scheduler', action_runner=mock.Mock())
        assert_not_equal(first, second)


class TestJobScheduler(TestCase):
    @setup
    def setup_job(self):
        mock_graph = mock.Mock(autospec=True)
        mock_graph.get_action_map.return_value = {}
        mock_graph.action_map = {}
        self.job = mock.Mock(autospec=True)
        self.job.allow_overlap = False
        self.job.max_runtime = datetime.timedelta(days=1)
        self.job_scheduler = JobScheduler(job=self.job)

    def test_restore_state_sets_job_runs(self):
        self.job.enabled = False
        mock_runs = [mock.Mock(), mock.Mock()]
        mock_action_runner = mock.Mock()
        job_state_data = {'runs': mock_runs, 'enabled': True}

        self.job_scheduler._set_callback = lambda x: x

        self.job.runs.runs = collections.deque()
        self.job.runs.get_scheduled.return_value = [mock.Mock()]
        self.job.get_job_runs_from_state.return_value = mock_runs

        with mock.patch(
            'tron.core.job_scheduler.recovery.launch_recovery_actionruns_for_job_runs',
            autospec=True,
        ) as mock_launch_recovery:
            mock_launch_recovery.return_value = mock.Mock(autospec=True)
            self.job_scheduler.restore_state(
                job_state_data, mock_action_runner
            )
            assert self.job.runs.runs == collections.deque(mock_runs)
            mock_launch_recovery.assert_called_once_with(
                job_runs=mock_runs, master_action_runner=mock_action_runner
            )
            calls = [mock.call(mock_runs[i]) for i in range(0, len(mock_runs))]
            self.job.watch.assert_has_calls(calls)

    def test_create_and_schedule_runs_specific_time(self):
        self.job_scheduler.get_runs_to_schedule = mock.Mock(return_value=[mock.Mock()])
        self.job_scheduler._set_callback = mock.Mock()
        self.job_scheduler.create_and_schedule_runs(next_run_time='a_datetime')
        assert self.job_scheduler.get_runs_to_schedule.call_args_list == [mock.call('a_datetime')]

    def test_create_and_schedule_runs_guess(self):
        self.job_scheduler.get_runs_to_schedule = mock.Mock(return_value=[mock.Mock()])
        self.job_scheduler._set_callback = mock.Mock()
        self.job_scheduler.create_and_schedule_runs(next_run_time=None)
        assert self.job_scheduler.get_runs_to_schedule.call_args_list == [mock.call(None)]

    def test_disable(self):
        self.job_scheduler.disable()
        assert self.job_scheduler.job.enabled is False
        self.job_scheduler.job.runs.cancel_pending.assert_called_once()

    def test_schedule_reconfigured(self):
        pending_run = mock.Mock()
        pending_run.run_time = 'a_run_time'
        self.job.runs.get_pending.return_value = [pending_run]
        self.job_scheduler.create_and_schedule_runs = mock.Mock()

        self.job_scheduler.schedule_reconfigured()

        assert self.job.runs.remove_pending.call_count == 1
        assert self.job_scheduler.create_and_schedule_runs.call_args_list == [mock.call(
            next_run_time='a_run_time',
        )]

    def test_schedule(self):
        self.job.enabled = True
        last_run = mock.Mock()
        last_run.run_time = 'a_run_time'
        self.job.runs.get_newest = mock.Mock(return_value=last_run)
        self.job_scheduler.create_and_schedule_runs = mock.Mock()

        self.job_scheduler.schedule()

        self.job.scheduler.next_run_time.assert_called_once_with('a_run_time')
        assert self.job_scheduler.create_and_schedule_runs.call_args_list == [
            mock.call(next_run_time=self.job.scheduler.next_run_time.return_value),
        ]

    def test_run_job(self):
        self.job_scheduler.schedule = mock.Mock(autospec=True)
        self.job.scheduler.schedule_on_complete = False
        self.job.runs.get_active = lambda n: []
        job_run = mock.Mock(autospec=True)
        job_run.is_cancelled = False
        self.job_scheduler.run_job(job_run)
        job_run.start.assert_called_once()
        self.job_scheduler.schedule.assert_called_once()

    def test_run_job_job_disabled(self):
        self.job_scheduler.schedule = MagicMock()
        job_run = MagicMock()
        self.job.enabled = False
        self.job_scheduler.run_job(job_run)
        assert_length(self.job_scheduler.schedule.mock_calls, 0)
        assert_length(job_run.start.mock_calls, 0)
        assert_length(job_run.cancel.mock_calls, 1)

    def test_run_job_cancelled(self):
        self.job_scheduler.schedule = MagicMock()
        job_run = MagicMock(is_scheduled=False)
        self.job_scheduler.run_job(job_run)
        assert_length(job_run.start.mock_calls, 0)
        assert_length(self.job_scheduler.schedule.mock_calls, 1)

    def test_run_job_already_running_queuing(self):
        self.job_scheduler.schedule = mock.Mock(autospec=True)
        self.job.runs.get_active = lambda s: [mock.Mock(autospec=True)]
        job_run = mock.Mock(autospec=True)
        job_run.is_cancelled = False
        self.job_scheduler.run_job(job_run)
        assert not job_run.start.called
        job_run.queue.assert_called_once()
        assert not self.job_scheduler.schedule.called

    def test_run_job_already_running_cancel(self):
        self.job_scheduler.schedule = mock.Mock(autospec=True)
        self.job.runs.get_active = lambda s: [mock.Mock(autospec=True)]
        self.job.queueing = False
        job_run = mock.Mock(autospec=True)
        job_run.is_cancelled = False
        self.job_scheduler.run_job(job_run)
        assert not job_run.start.called
        job_run.cancel.assert_called_once()
        self.job_scheduler.schedule.assert_called_once()

    def test_run_job_already_running_allow_overlap(self):
        self.job_scheduler.schedule = mock.Mock()
        self.job.runs.get_active = lambda s: [mock.Mock()]
        self.job.allow_overlap = True
        job_run = MagicMock(is_cancelled=False)
        self.job_scheduler.run_job(job_run)
        job_run.start.assert_called_with()

    def test_run_job_has_starting_queueing(self):
        self.job_scheduler.schedule = mock.Mock(autospec=True)
        self.job.runs.get_active = lambda s: [mock.Mock(autospec=True)]
        job_run = mock.Mock(autospec=True)
        job_run.is_cancelled = False
        self.job_scheduler.run_job(job_run)
        assert not job_run.start.called
        job_run.queue.assert_called_once()
        assert not self.job_scheduler.schedule.called

    def test_run_job_schedule_on_complete(self):
        self.job_scheduler.schedule = MagicMock()
        self.job.scheduler.schedule_on_complete = True
        self.job.runs.get_active = lambda s: []
        job_run = MagicMock(is_cancelled=False)
        self.job_scheduler.run_job(job_run)
        assert_length(job_run.start.mock_calls, 1)
        assert_length(self.job_scheduler.schedule.mock_calls, 0)
